# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
import numbers
import json

import jax
import jax.tree_util

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


def _find_dataptrs(layout):
    def find_nparray_ptrs(node, depth, data_ptrs):
        if isinstance(node, ak.layout.NumpyArray):
            data_ptrs.append(node.ptr)

    data_ptrs = []
    ak._util.recursive_walk(layout, find_nparray_ptrs, args=(data_ptrs,))

    return data_ptrs


def _find_dataptrs_and_map(layout, jaxtracers, isscalar):
    if not isscalar:
        data_ptrs = _find_dataptrs(layout)
        assert len(jaxtracers) == len(data_ptrs)
        map_ptrs_to_tracers = dict(zip(data_ptrs, jaxtracers))
    else:
        # Layout is a scalar
        data_ptrs = None
        map_ptrs_to_tracers = None

    return data_ptrs, map_ptrs_to_tracers


def _jaxtracers_getitem(array, where):
    if array._isscalar:
        raise TypeError("Cannot slice a scalar")

    out = array.layout[where]

    def find_nparray_node_newptr(layout, outlayout):
        def find_nparray_node(
            node, depth, outlayout, fieldloc, shape, nodenum, nodenum_index
        ):
            if isinstance(node, ak.layout.NumpyArray):
                if node.identities.fieldloc == fieldloc:
                    nodenum_index.append(nodenum)
                else:
                    nodenum = nodenum + 1

        outlayout_fieldloc = outlayout.identities.fieldloc
        nodenum_index = []

        ak._util.recursive_walk(
            layout,
            find_nparray_node,
            args=(
                outlayout,
                outlayout_fieldloc,
                numpy.asarray(outlayout.identities).shape,
                0,
                nodenum_index,
            ),
        )
        if len(nodenum_index) == 0:
            raise ValueError("Couldn't find the node in new slice")
        return nodenum_index[0]

    if not isinstance(out, ak.layout.Content):

        def recurse(outarray, recurse_where):

            if isinstance(recurse_where, numbers.Integral) or isinstance(
                recurse_where, str
            ):
                if isinstance(outarray.layout, ak.layout.NumpyArray):
                    return outarray._tracers[0][recurse_where]
            elif isinstance(where, tuple):
                return recurse(array[where[:-1]], where[len(where) - 1])

            else:
                raise ValueError("Can't slice the array with {0}".format(where))

        child = [recurse(array, where)]
        return ak.Array._internal_for_jax(out, child, isscalar=True)

    else:

        def fetch_indices_and_fieldloc_layout(outlayout):
            if isinstance(outlayout, ak.layout.NumpyArray):
                return [
                    (
                        (
                            outlayout.identities.fieldloc,
                            numpy.asarray(outlayout.identities).shape[1],
                        ),
                        numpy.asarray(outlayout.identities),
                    )
                ]
            elif isinstance(outlayout, ak._util.listtypes):
                return fetch_indices_and_fieldloc_layout(outlayout.content)
            elif isinstance(outlayout, ak._util.indexedtypes):
                return fetch_indices_and_fieldloc_layout(outlayout.project())
            elif isinstance(outlayout, ak._util.uniontypes):
                raise ValueError(
                    "Can't differentiate an UnionArray type {0}".format(outlayout)
                )
            elif isinstance(outlayout, ak._util.recordtypes):
                indices = []
                for content in outlayout.contents:
                    indices = indices + fetch_indices_and_fieldloc_layout(content)
                return indices
            elif isinstance(outlayout, ak._util.indexedtypes):
                return fetch_indices_and_fieldloc_layout(outlayout.content)
            elif isinstance(outlayout, ak._util.indexedoptiontypes):
                return fetch_indices_and_fieldloc_layout(outlayout.content)
            elif isinstance(
                outlayout,
                (
                    ak.layout.BitMaskedArray,
                    ak.layout.ByteMaskedArray,
                    ak.layout.UnmaskedArray,
                ),
            ):
                return fetch_indices_and_fieldloc_layout(outlayout.content)
            else:
                raise NotImplementedError

        def fetch_children_tracer(outlayout, preslice_identities):
            if isinstance(outlayout, ak.layout.NumpyArray):

                def find_intersection_indices(
                    postslice_identities, preslice_identities
                ):
                    multiplier = numpy.append(
                        numpy.cumprod(
                            (numpy.max(preslice_identities, axis=0) + 1)[::-1]
                        )[-2::-1],
                        1,
                    )
                    haystack = numpy.sum(preslice_identities * multiplier, axis=1)
                    needle = numpy.sum(postslice_identities * multiplier, axis=1)
                    haystack_argsort = numpy.argsort(haystack)
                    indices = numpy.searchsorted(
                        haystack, needle, sorter=haystack_argsort
                    )
                    final_indices = haystack_argsort[indices]
                    return final_indices

                def find_corresponding_identity(
                    postslice_identities, preslice_identities
                ):
                    for identity in preslice_identities:
                        if identity[0] == postslice_identities:
                            return identity[1]

                    raise ValueError(
                        "Couldn't find postslice identities in preslice identities"
                    )

                indices = find_intersection_indices(
                    numpy.asarray(outlayout.identities),
                    find_corresponding_identity(
                        (
                            outlayout.identities.fieldloc,
                            numpy.asarray(outlayout.identities).shape[1],
                        ),
                        preslice_identities,
                    ),
                )
                if outlayout.ptr in array._map_ptrs_to_tracers:
                    tracer = array._map_ptrs_to_tracers[outlayout.ptr]
                else:
                    tracer = array._tracers[
                        find_nparray_node_newptr(array.layout, outlayout)
                    ]
                return [jax.numpy.take(tracer, indices)]

            elif isinstance(outlayout, ak._util.listtypes):
                return fetch_children_tracer(outlayout.content, preslice_identities)
            elif isinstance(outlayout, ak._util.uniontypes):
                raise ValueError(
                    "Can't differentiate an UnionArray type {0}".format(outlayout)
                )
            elif isinstance(outlayout, ak._util.recordtypes):
                children = []
                for content in outlayout.contents:
                    children = children + fetch_children_tracer(
                        content, preslice_identities
                    )
                return children
            elif isinstance(outlayout, ak._util.indexedtypes):
                return fetch_children_tracer(outlayout.content, preslice_identities)
            elif isinstance(outlayout, ak._util.indexedoptiontypes):
                return fetch_children_tracer(outlayout.content, preslice_identities)
            elif isinstance(
                outlayout,
                (
                    ak.layout.BitMaskedArray,
                    ak.layout.ByteMaskedArray,
                    ak.layout.UnmaskedArray,
                ),
            ):
                return fetch_children_tracer(outlayout.content, preslice_identities)
            else:
                raise NotImplementedError(
                    "fetch_children_tracer not completely implemented yet for {0}".format(
                        outlayout
                    )
                )

        children = fetch_children_tracer(
            out, fetch_indices_and_fieldloc_layout(array.layout)
        )
        out = out.deep_copy()
        out.setidentities()
        return ak.Array._internal_for_jax(out, children, isscalar=False)


def array_ufunc(array, ufunc, method, inputs, kwargs):
    for x in inputs:
        if isinstance(x, ak.Array) and hasattr(x, "_tracers"):
            assert len(x._tracers) == len(array._tracers)

    # ak.Array __add__, etc. map to the NumPy functions, switch to JAX
    import numpy

    for name, np_ufunc in numpy.core.umath.__dict__.items():
        if ufunc is np_ufunc:
            ufunc = getattr(jax.numpy, name)

    # need to apply the ufunc to the same argument list for each tracer separately
    nexttracers = []
    for i in range(len(array._tracers)):
        nextinputs = [
            x._tracers[i] if isinstance(x, ak.Array) and hasattr(x, "_tracers") else x
            for x in inputs
        ]
        nexttracers.append(getattr(ufunc, method)(*nextinputs, **kwargs))

    return ak.Array._internal_for_jax(array.layout, nexttracers, array._isscalar)


class AuxData(object):
    def __init__(self, array):
        self.array = array

    def __eq__(self, other):
        self_form = json.loads(self.array.layout.form.tojson())
        other_form = json.loads(self.array.layout.form.tojson())

        def form_sweep(input_form, blacklist):
            if isinstance(input_form, dict):
                return {
                    k: form_sweep(v, blacklist)
                    for k, v in input_form.items()
                    if k not in blacklist
                }
            else:
                return input_form

        self_form = form_sweep(self_form, ["primitive", "format", "itemsize"])
        other_form = form_sweep(self_form, ["primitive", "format", "itemsize"])

        return self_form == other_form


def special_flatten(array):
    if isinstance(array, ak.Array) and hasattr(array, "_tracers"):
        aux_data, children = AuxData(array), array._tracers
    elif isinstance(array, ak.Array):

        def create_databuffers(node, depth, databuffers):
            if isinstance(node, ak.layout.NumpyArray):
                databuffers.append(node)

        databuffers = []
        ak._util.recursive_walk(array.layout, create_databuffers, args=(databuffers,))

        array.layout.setidentities()
        children = [jax.numpy.asarray(x) for x in databuffers]
        aux_data = AuxData(ak.Array._internal_for_jax(array.layout, children))
    else:
        raise ValueError(
            "Can only differentiate Awkward Arrays, received array of type {0}".format(
                type(array)
            )
        )

    return children, aux_data


def special_unflatten(aux_data, children):
    if any(isinstance(x, jax.core.Tracer) for x in children):
        return ak.Array._internal_for_jax(aux_data.array.layout, children)
    elif all(child is None for child in children):
        return None
    else:
        if aux_data.array._isscalar:
            assert len(children) == 1
            # return children[0]
            import numpy

            return numpy.ndarray.item(numpy.asarray(children[0]))
        children = list(children)

        def function(layout):
            if isinstance(layout, ak.layout.NumpyArray):
                buffer = children[0]
                children.pop(0)
                return lambda: ak.layout.NumpyArray(buffer)

        arr = ak._util.recursively_apply(
            aux_data.array.layout, function, pass_depth=False
        )
        return ak.Array(arr)


jax.tree_util.register_pytree_node(ak.Array, special_flatten, special_unflatten)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import awkward as ak
from awkward._v2.contents.content import Content

np = ak.nplike.NumpyMetadata.instance()


class Record(object):
    def __init__(self, array, at):
        if not isinstance(array, ak._v2.contents.recordarray.RecordArray):
            raise TypeError(
                "Record 'array' must be a RecordArray, not {0}".format(repr(array))
            )
        if not ak._util.isint(at):
            raise TypeError(
                "Record 'at' must be an integer, not {0}".format(repr(array))
            )
        if 0 <= at < len(array):
            self._array = array
            self._at = at
        else:
            raise ValueError(
                "Record 'at' must be >= 0 and < len(array) == {0}, not {1}".format(
                    len(array), at
                )
            )

    @property
    def array(self):
        return self._array

    @property
    def at(self):
        return self._at

    @property
    def keys(self):
        return self._array.keys

    @property
    def is_tuple(self):
        return self._array.keys is None

    @property
    def as_tuple(self):
        return Record(self._array.as_tuple, self._at)

    @property
    def contents(self):
        out = []
        for key in self._array.keys:
            out.append(self._array[key][self._at])
        return out

    def content(self, index_or_key):
        return self._array.content(index_or_key)[self._at]

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<Record at="]
        out.append(repr(str(self._at)))
        out.append(">\n")
        out.append(self._array._repr(indent + "    ", "<array>", "</array>\n"))
        out.append(indent)
        out.append("</Record>")
        out.append(post)
        return "".join(out)

    def __getitem__(self, where):
        if ak._util.isint(where):
            raise IndexError("scalar Record cannot be sliced by an integer")

        elif isinstance(where, slice):
            raise IndexError("scalar Record cannot be sliced by a range slice (`:`)")

        elif ak._util.isstr(where):
            return self._getitem_field(where)

        elif where is np.newaxis:
            raise IndexError("scalar Record cannot be sliced by np.newaxis (`None`)")

        elif where is Ellipsis:
            raise IndexError("scalar Record cannot be sliced by an ellipsis (`...`)")

        elif isinstance(where, tuple) and len(where) == 0:
            return self

        elif isinstance(where, tuple) and len(where) == 1:
            return self.__getitem__(where[0])

        elif isinstance(where, tuple) and ak._util.isstr(where[0]):
            return self._getitem_field(where[0]).__getitem__(where[1:])

        elif isinstance(where, ak.highlevel.Array):
            raise IndexError("scalar Record cannot be sliced by an array")

        elif isinstance(where, ak.layout.Content):
            raise IndexError("scalar Record cannot be sliced by an array")

        elif isinstance(where, Content):
            raise IndexError("scalar Record cannot be sliced by an array")

        elif isinstance(where, Iterable) and all(ak._util.isstr(x) for x in where):
            return self._getitem_fields(where)

        elif isinstance(where, Iterable):
            raise IndexError("scalar Record cannot be sliced by an array")

        else:
            raise TypeError(
                "only field name (str) or names (non-tuple iterable of str) "
                "are valid indices for slicing a scalar record, not\n\n    "
                + repr(where)
            )

    def _getitem_field(self, where):
        return self._array._getitem_field(where)._getitem_at(self._at)

    def _getitem_fields(self, where):
        return self._array._getitem_fields(where)._getitem_at(self._at)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


@pytest.mark.skipif(not ak._util.win, reason="need Windows to test this case")
def test():
    if not ak._util.py27:
        # Python 2.7 has differently named urllib libraries
        import urllib.request
        import urllib.error

        url = "https://raw.githubusercontent.com/Chicago/osd-bike-routes/5f556dc/data/Bikeroutes.geojson"
        try:
            bikeroutes_json = urllib.request.urlopen(url).read()
        except urllib.error.URLError:
            pytest.skip(msg="couldn't download sample dataset")

        # This shouldn't fail (see #1084)
        ak.from_json(bikeroutes_json)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
from awkward._v2.forms.form import (
    Form,
    _parameters_equal,
    _parameters_update,
    nonvirtual,
)


class VirtualForm(Form):
    def __init__(
        self,
        form,
        has_length,
        has_identifier=False,
        parameters=None,
        form_key=None,
    ):
        if form is not None and not isinstance(form, Form):
            raise TypeError(
                "{0} 'form' must be a Form instance, not {1}".format(
                    type(self).__name__, repr(form)
                )
            )
        if not isinstance(has_length, bool):
            raise TypeError(
                "{0} 'has_length' must be bool, not {1}".format(
                    type(self).__name__, repr(has_length)
                )
            )

        self._form = form
        self._has_length = has_length
        self._init(has_identifier, parameters, form_key)

    @property
    def form(self):
        return self._form

    @property
    def has_length(self):
        return self._has_length

    def __repr__(self):
        args = [repr(self._form), repr(self._has_length)] + self._repr_args()
        return "{0}({1})".format(type(self).__name__, ", ".join(args))

    def _tolist_part(self, verbose, toplevel):
        out = {"class": "VirtualArray"}
        if self._form is None:
            out["form"] = None
        else:
            out["form"] = self._form._tolist_part(verbose, toplevel=False)
        out["has_length"] = self._has_length
        return self._tolist_extra(out, verbose)

    def _type(self, typestrs):
        if self._form is None:
            return ak._v2.types.unknowntype.UnknownType(
                self._parameters,
                ak._util.gettypestr(self._parameters, typestrs),
            )
        else:
            out = self._form._type(typestrs)

            if self._parameters is not None:
                if out._parameters is None:
                    out._parameters = self._parameters
                else:
                    out._parameters = dict(out._parameters)
                    _parameters_update(out._parameters, self._parameters)

            return out

    def __eq__(self, other):
        if isinstance(other, VirtualForm):
            return (
                self._has_identifier == other._has_identifier
                and self._form_key == other._form_key
                and _parameters_equal(self._parameters, other._parameters)
                and self._form == other._form
            )
        else:
            return False

    def generated_compatibility(self, other):
        if other is None:
            other_parameters = None
        else:
            other_parameters = other._parameters

        if not _parameters_equal(self._parameters, other_parameters):
            return False

        nonvirtual_self = nonvirtual(self)
        nonvirtual_other = nonvirtual(other)

        if nonvirtual_self is None or nonvirtual_other is None:
            return True
        else:
            return nonvirtual_self.generated_compatibility(nonvirtual_other)

    def _getitem_range(self):
        return VirtualForm(
            None if self._form is None else self._form._getitem_range(),
            True,
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    def _getitem_field(self, where, only_fields=()):
        return VirtualForm(
            None
            if self._form is None
            else self._form._getitem_field(where, only_fields),
            self._has_length,
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return VirtualForm(
            None
            if self._form is None
            else self._form._getitem_fields(where, only_fields),
            self._has_length,
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _carry(self, allow_lazy):
        return VirtualForm(
            None if self._form is None else self._form._getitem_range(),
            True,
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def purelist_isregular(self):
        if self._form is None:
            raise ValueError(
                "cannot determine the type of a virtual array without a Form"
            )
        else:
            return self._form.purelist_isregular

    @property
    def purelist_depth(self):
        if self._form is None:
            raise ValueError(
                "cannot determine the type of a virtual array without a Form"
            )
        else:
            return self._form.purelist_depth

    @property
    def minmax_depth(self):
        if self._form is None:
            raise ValueError(
                "cannot determine the type of a virtual array without a Form"
            )
        else:
            return self._form.minmax_depth

    @property
    def branch_depth(self):
        if self._form is None:
            raise ValueError(
                "cannot determine the type of a virtual array without a Form"
            )
        else:
            return self._form.branch_depth

    @property
    def keys(self):
        if self._form is None:
            raise ValueError(
                "cannot determine the type of a virtual array without a Form"
            )
        else:
            return self._form.keys

    @property
    def dimension_optiontype(self):
        if self._form is None:
            raise ValueError(
                "cannot determine the type of a virtual array without a Form"
            )
        else:
            return self._form.dimension_optiontype



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import os.path

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


numba = pytest.importorskip("numba")


def test_constructed():
    @numba.njit
    def f1(array):
        out = np.ones(6, np.float64)
        i = 0
        for x in array:
            out[i] = x
            i += 1
        return out

    content = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5]))
    virtual1 = ak.layout.VirtualArray(
        ak.layout.ArrayGenerator(
            lambda: content, form=content.form, length=len(content)
        )
    )
    virtual2 = ak.layout.VirtualArray(
        ak.layout.ArrayGenerator(
            lambda: virtual1, form=virtual1.form, length=len(virtual1)
        )
    )
    array = ak.Array(virtual2)

    assert f1(array).tolist() == [0.0, 1.1, 2.2, 3.3, 4.4, 5.5]


def test_enumerate_Partitioned():
    @numba.njit
    def f1(array):
        for _, _ in enumerate(array):
            pass

    array = ak.repartition(ak.Array(range(100)), 10)
    f1(array)


pytest.importorskip("pyarrow.parquet")


def test_parquet1(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")
    array = ak.Array([{"x": {"y": 0.0}}, {"x": {"y": 1.1}}, {"x": {"y": 2.2}}])
    ak.to_parquet(array, filename)

    lazy = ak.from_parquet(filename, lazy=True, lazy_cache=None)

    @numba.njit
    def f1(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for obj in lazy:
            out[i] = obj.x.y
            i += 1
        return out

    assert f1(lazy).tolist() == [0.0, 1.1, 2.2]


def test_parquet2(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")
    array = ak.Array([{"x": [{"y": 0.0}]}, {"x": [{"y": 1.1}]}, {"x": [{"y": 2.2}]}])
    ak.to_parquet(array, filename)

    lazy = ak.from_parquet(filename, lazy=True, lazy_cache=None)

    @numba.njit
    def f1(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for obj in lazy:
            for subobj in obj.x:
                out[i] = subobj.y
                i += 1
        return out

    assert f1(lazy).tolist() == [0.0, 1.1, 2.2]


def test_parquet3(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")
    array = ak.Array(
        [[{"x": [{"y": 0.0}]}], [{"x": [{"y": 1.1}]}], [{"x": [{"y": 2.2}]}]]
    )
    ak.to_parquet(array, filename)

    lazy = ak.from_parquet(filename, lazy=True, lazy_cache=None)

    @numba.njit
    def f1(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for sublist in lazy:
            for obj in sublist:
                for subobj in obj.x:
                    out[i] = subobj.y
                    i += 1
        return out

    assert f1(lazy).tolist() == [0.0, 1.1, 2.2]


def test_parquet4(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")
    array = ak.Array(
        [
            {"z": [{"x": [{"y": 0.0}]}]},
            {"z": [{"x": [{"y": 1.1}]}]},
            {"z": [{"x": [{"y": 2.2}]}]},
        ]
    )
    ak.to_parquet(array, filename)

    lazy = ak.from_parquet(filename, lazy=True, lazy_cache=None)

    @numba.njit
    def f1(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for sublist in lazy:
            for obj in sublist.z:
                for subobj in obj.x:
                    out[i] = subobj.y
                    i += 1
        return out

    assert f1(lazy).tolist() == [0.0, 1.1, 2.2]


def test_parquet1b(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")
    array = ak.Array(
        [
            {"x": {"y": 0.0, "z": 0}},
            {"x": {"y": 1.1, "z": 1}},
            {"x": {"y": 2.2, "z": 2}},
        ]
    )
    ak.to_parquet(array, filename)

    lazy = ak.from_parquet(filename, lazy=True, lazy_cache=None)

    @numba.njit
    def f1(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for obj in lazy:
            out[i] = obj.x.y
            i += 1
        return out

    @numba.njit
    def f2(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for obj in lazy:
            out[i] = obj.x.z
            i += 1
        return out

    assert f1(lazy).tolist() == [0.0, 1.1, 2.2]
    assert f2(lazy).tolist() == [0, 1, 2]


def test_parquet2b(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")
    array = ak.Array(
        [
            {"x": [{"y": 0.0, "z": 0}]},
            {"x": [{"y": 1.1, "z": 1}]},
            {"x": [{"y": 2.2, "z": 2}]},
        ]
    )
    ak.to_parquet(array, filename)

    lazy = ak.from_parquet(filename, lazy=True, lazy_cache=None)

    @numba.njit
    def f1(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for obj in lazy:
            for subobj in obj.x:
                out[i] = subobj.y
                i += 1
        return out

    @numba.njit
    def f2(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for obj in lazy:
            for subobj in obj.x:
                out[i] = subobj.z
                i += 1
        return out

    assert f1(lazy).tolist() == [0.0, 1.1, 2.2]
    assert f2(lazy).tolist() == [0, 1, 2]


def test_parquet3b(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")
    array = ak.Array(
        [
            [{"x": [{"y": 0.0, "z": 0}]}],
            [{"x": [{"y": 1.1, "z": 1}]}],
            [{"x": [{"y": 2.2, "z": 2}]}],
        ]
    )
    ak.to_parquet(array, filename)

    lazy = ak.from_parquet(filename, lazy=True, lazy_cache=None)

    @numba.njit
    def f1(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for sublist in lazy:
            for obj in sublist:
                for subobj in obj.x:
                    out[i] = subobj.y
                    i += 1
        return out

    @numba.njit
    def f2(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for sublist in lazy:
            for obj in sublist:
                for subobj in obj.x:
                    out[i] = subobj.z
                    i += 1
        return out

    assert f1(lazy).tolist() == [0.0, 1.1, 2.2]
    assert f2(lazy).tolist() == [0, 1, 2]


def test_parquet4b(tmp_path):
    filename = os.path.join(tmp_path, "whatever.parquet")
    array = ak.Array(
        [
            {"z": [{"x": 0.0, "w": 0}]},
            {"z": [{"x": 1.1, "w": 1}]},
            {"z": [{"x": 2.2, "w": 2}]},
        ]
    )
    ak.to_parquet(array, filename)

    lazy = ak.from_parquet(filename, lazy=True, lazy_cache=None)

    @numba.njit
    def f1(lazy):
        out = np.ones(3, np.float64)
        i = 0
        for sublist in lazy:
            for obj in sublist.z:
                out[i] = obj.x
                i += 1
        return out

    assert f1(lazy).tolist() == [0.0, 1.1, 2.2]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_nan():
    array = ak.Array([1, 2, np.nan, 3, 0, np.nan])

    assert ak.argsort(array).tolist() == [
        2,
        5,
        4,
        0,
        1,
        3,
    ]
    # Note, `nan` comparison with `nan` returns False
    assert str(ak.sort(array).tolist()) == "[nan, nan, 0.0, 1.0, 2.0, 3.0]"


def test_bool():
    array = ak.Array([True, False, False, True, True, True])

    assert ak.argsort(array).tolist() == [1, 2, 0, 3, 4, 5]
    assert ak.sort(array).tolist() == [False, False, True, True, True, True]


def test_argsort():
    array = ak.Array([1, 2, None, 3, 0, None])
    assert ak.argsort(array).tolist() == [4, 0, 1, 3, 2, 5]
    assert array[ak.argsort(array)].tolist() == [
        0,
        1,
        2,
        3,
        None,
        None,
    ]


def test_argsort_2d():
    array = ak.Array([[1, 2, None, 3, 0, None], [1, 2, None, 3, 0, None]])
    assert ak.argsort(array).tolist() == [[4, 0, 1, 3, 2, 5], [4, 0, 1, 3, 2, 5]]
    assert array[ak.argsort(array)].tolist() == [
        [
            0,
            1,
            2,
            3,
            None,
            None,
        ],
        [
            0,
            1,
            2,
            3,
            None,
            None,
        ],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array([[1, 2, 3], [], [4, 5]])
    assert array[2].tolist() == [4, 5]
    assert (
        array[
            2,
        ].tolist()
        == [4, 5]
    )
    assert array[2, 1] == 5

    assert array[np.int64(2)].tolist() == [4, 5]
    assert (
        array[
            np.int64(2),
        ].tolist()
        == [4, 5]
    )
    assert array[np.int64(2), 1] == 5
    assert array[2, np.int64(1)] == 5
    assert array[np.int64(2), np.int64(1)] == 5

    assert array[np.int64([2]).reshape(())].tolist() == [4, 5]
    assert (
        array[
            np.int64([2]).reshape(()),
        ].tolist()
        == [4, 5]
    )
    assert array[np.int64([2]).reshape(()), 1] == 5
    assert array[2, np.int64([1]).reshape(())] == 5
    assert array[np.int64([2]).reshape(()), np.int64([1]).reshape(())] == 5



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import json

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_basic():
    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5], dtype=np.int64))
    content2 = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], dtype=np.float64)
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content2)
    recordarray = ak.layout.RecordArray(
        [content1, listoffsetarray, content2, content1],
        keys=["one", "two", "2", "wonky"],
    )

    recordarray = v1_to_v2(recordarray)

    assert ak.to_list(recordarray.content(0)) == [1, 2, 3, 4, 5]
    assert recordarray.typetracer.content(0).form == recordarray.content(0).form
    assert ak.to_list(recordarray.content("two")) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]
    assert recordarray.typetracer.content("two").form == recordarray.content("two").form
    assert ak.to_list(recordarray.content("wonky")) == [1, 2, 3, 4, 5]
    assert (
        recordarray.typetracer.content("wonky").form
        == recordarray.content("wonky").form
    )

    str(recordarray)

    assert json.loads(ak._v2.forms.form.Form.to_json(recordarray.form)) == (
        {
            "class": "RecordArray",
            "contents": {
                "one": {
                    "class": "NumpyArray",
                    "primitive": "int64",
                    "inner_shape": [],
                    "has_identifier": False,
                    "parameters": {},
                    "form_key": None,
                },
                "two": {
                    "class": "ListOffsetArray",
                    "offsets": "i64",
                    "content": {
                        "class": "NumpyArray",
                        "primitive": "float64",
                        "inner_shape": [],
                        "has_identifier": False,
                        "parameters": {},
                        "form_key": None,
                    },
                    "has_identifier": False,
                    "parameters": {},
                    "form_key": None,
                },
                "2": {
                    "class": "NumpyArray",
                    "primitive": "float64",
                    "inner_shape": [],
                    "has_identifier": False,
                    "parameters": {},
                    "form_key": None,
                },
                "wonky": {
                    "class": "NumpyArray",
                    "primitive": "int64",
                    "inner_shape": [],
                    "has_identifier": False,
                    "parameters": {},
                    "form_key": None,
                },
            },
            "has_identifier": False,
            "parameters": {},
            "form_key": None,
        }
    )

    assert len(recordarray) == 5
    assert recordarray.index_to_key(0) == "one"
    assert recordarray.index_to_key(1) == "two"
    assert recordarray.index_to_key(2) == "2"
    assert recordarray.index_to_key(3) == "wonky"
    assert recordarray.key_to_index("wonky") == 3
    assert recordarray.key_to_index("one") == 0
    # FIXME
    # assert recordarray.key_to_index("0") == 0
    assert recordarray.key_to_index("two") == 1
    # assert recordarray.key_to_index("1") == 1
    assert recordarray.key_to_index("2") == 2
    assert recordarray.haskey("wonky")
    assert recordarray.haskey("one")
    # assert recordarray.haskey("0")
    assert recordarray.haskey("two")
    # assert recordarray.haskey("1")
    assert recordarray.haskey("2")

    assert recordarray.keys == ["one", "two", "2", "wonky"]
    assert [ak.to_list(x) for x in recordarray.contents] == [
        [1, 2, 3, 4, 5],
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]],
        [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9],
        [1, 2, 3, 4, 5],
    ]

    pairs = list(zip(recordarray.keys, recordarray.contents))
    assert pairs[0][0] == "one"
    assert pairs[1][0] == "two"
    assert pairs[2][0] == "2"
    assert pairs[3][0] == "wonky"
    assert ak.to_list(pairs[0][1]) == [1, 2, 3, 4, 5]
    assert ak.to_list(pairs[1][1]) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]
    assert ak.to_list(pairs[2][1]) == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]
    assert ak.to_list(pairs[3][1]) == [1, 2, 3, 4, 5]

    assert json.loads(ak._v2.forms.form.Form.to_json(recordarray.form)) == {
        "class": "RecordArray",
        "contents": {
            "one": {
                "class": "NumpyArray",
                "primitive": "int64",
                "inner_shape": [],
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
            "two": {
                "class": "ListOffsetArray",
                "offsets": "i64",
                "content": {
                    "class": "NumpyArray",
                    "primitive": "float64",
                    "inner_shape": [],
                    "has_identifier": False,
                    "parameters": {},
                    "form_key": None,
                },
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
            "2": {
                "class": "NumpyArray",
                "primitive": "float64",
                "inner_shape": [],
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
            "wonky": {
                "class": "NumpyArray",
                "primitive": "int64",
                "inner_shape": [],
                "has_identifier": False,
                "parameters": {},
                "form_key": None,
            },
        },
        "has_identifier": False,
        "parameters": {},
        "form_key": None,
    }
    assert json.loads(ak.to_json(recordarray.as_tuple)) == [
        {"0": 1, "1": [1.1, 2.2, 3.3], "2": 1.1, "3": 1},
        {"0": 2, "1": [], "2": 2.2, "3": 2},
        {"0": 3, "1": [4.4, 5.5], "2": 3.3, "3": 3},
        {"0": 4, "1": [6.6], "2": 4.4, "3": 4},
        {"0": 5, "1": [7.7, 8.8, 9.9], "2": 5.5, "3": 5},
    ]


def test_scalar_record():
    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5], dtype=np.int64))
    content2 = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], dtype=np.float64)
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content2)
    recordarray = ak.layout.RecordArray(
        [content1, listoffsetarray], keys=["one", "two"]
    )

    listoffsetarray = v1_to_v2(listoffsetarray)
    recordarray = v1_to_v2(recordarray)

    str(recordarray)
    str(recordarray[2])
    assert recordarray.typetracer[2].array.form == recordarray[2].array.form

    assert recordarray[2].keys == ["one", "two"]
    assert [ak.to_list(x) for x in recordarray[2].contents] == [3, [4.4, 5.5]]
    pairs = [(key, recordarray[2].content(key)) for key in recordarray[2].keys]
    assert pairs[0][0] == "one"
    assert pairs[1][0] == "two"
    assert pairs[0][1] == 3
    assert ak.to_list(pairs[1][1]) == [4.4, 5.5]
    assert ak.to_list(recordarray[2]) == {"one": 3, "two": [4.4, 5.5]}


def test_getitem():
    assert (
        ak._ext._slice_tostring((1, 2, [3], "four", ["five", "six"], slice(7, 8, 9)))
        == '[array([1]), array([2]), array([3]), "four", ["five", "six"], 7:8:9]'
    )

    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5], dtype=np.int64))
    content2 = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], dtype=np.float64)
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content2)
    recordarray = ak.layout.RecordArray([content1, listoffsetarray, content2])
    recordarray = v1_to_v2(recordarray)

    assert ak.to_list(recordarray["2"]) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert recordarray.typetracer["2"].form == recordarray["2"].form
    assert ak.to_list(recordarray[["0", "1"]]) == [
        (1, [1.1, 2.2, 3.3]),
        (2, []),
        (3, [4.4, 5.5]),
        (4, [6.6]),
        (5, [7.7, 8.8, 9.9]),
    ]
    assert recordarray.typetracer[["0", "1"]].form == recordarray[["0", "1"]].form
    assert ak.to_list(recordarray[["1", "0"]]) == [
        ([1.1, 2.2, 3.3], 1),
        ([], 2),
        ([4.4, 5.5], 3),
        ([6.6], 4),
        ([7.7, 8.8, 9.9], 5),
    ]
    assert recordarray.typetracer[["1", "0"]].form == recordarray[["1", "0"]].form
    assert ak.to_list(recordarray[1:-1]) == [
        (2, [], 2.2),
        (3, [4.4, 5.5], 3.3),
        (4, [6.6], 4.4),
    ]
    assert recordarray.typetracer[1:-1].form == recordarray[1:-1].form
    assert ak.to_list(recordarray[2]) == (3, [4.4, 5.5], 3.3)
    assert recordarray.typetracer[2].array.form == recordarray[2].array.form
    assert ak.to_list(recordarray[2]["1"]) == [4.4, 5.5]
    assert recordarray.typetracer[2]["1"].form == recordarray[2]["1"].form
    assert ak.to_list(recordarray[2][["0", "1"]]) == (3, [4.4, 5.5])
    assert (
        recordarray.typetracer[2][["0", "1"]].array.form
        == recordarray[2][["0", "1"]].array.form
    )
    assert ak.to_list(recordarray[2][["1", "0"]]) == ([4.4, 5.5], 3)
    assert (
        recordarray.typetracer[2][["1", "0"]].array.form
        == recordarray[2][["1", "0"]].array.form
    )

    recordarray = ak.layout.RecordArray(
        {"one": content1, "two": listoffsetarray, "three": content2}
    )
    recordarray = v1_to_v2(recordarray)

    assert not recordarray.is_tuple

    assert ak.to_list(recordarray["three"]) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert recordarray.typetracer["three"].form == recordarray["three"].form
    assert ak.to_list(recordarray[["one", "two"]]) == [
        {"one": 1, "two": [1.1, 2.2, 3.3]},
        {"one": 2, "two": []},
        {"one": 3, "two": [4.4, 5.5]},
        {"one": 4, "two": [6.6]},
        {"one": 5, "two": [7.7, 8.8, 9.9]},
    ]
    assert (
        recordarray.typetracer[["one", "two"]].form == recordarray[["one", "two"]].form
    )
    assert ak.to_list(recordarray[["two", "one"]]) == [
        {"one": 1, "two": [1.1, 2.2, 3.3]},
        {"one": 2, "two": []},
        {"one": 3, "two": [4.4, 5.5]},
        {"one": 4, "two": [6.6]},
        {"one": 5, "two": [7.7, 8.8, 9.9]},
    ]
    assert (
        recordarray.typetracer[["two", "one"]].form == recordarray[["two", "one"]].form
    )
    assert ak.to_list(recordarray[1:-1]) == [
        {"one": 2, "two": [], "three": 2.2},
        {"one": 3, "two": [4.4, 5.5], "three": 3.3},
        {"one": 4, "two": [6.6], "three": 4.4},
    ]
    assert recordarray.typetracer[1:-1].form == recordarray[1:-1].form
    assert ak.to_list(recordarray[2]) == {"one": 3, "two": [4.4, 5.5], "three": 3.3}
    assert recordarray.typetracer[2].array.form == recordarray[2].array.form
    assert ak.to_list(recordarray[2]["two"]) == [4.4, 5.5]
    assert recordarray.typetracer[2]["two"].form == recordarray[2]["two"].form
    assert ak.to_list(recordarray[2][["one", "two"]]) == {"one": 3, "two": [4.4, 5.5]}
    assert (
        recordarray.typetracer[2][["one", "two"]].array.form
        == recordarray[2][["one", "two"]].array.form
    )
    assert ak.to_list(recordarray[2][["two", "one"]]) == {"one": 3, "two": [4.4, 5.5]}
    assert (
        recordarray.typetracer[2][["two", "one"]].array.form
        == recordarray[2][["two", "one"]].array.form
    )


def test_getitem_other_types():
    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5], dtype=np.int64))
    content2 = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], dtype=np.float64)
    )
    offsets1 = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9], dtype=np.int64))
    listoffsetarray1 = ak.layout.ListOffsetArray64(offsets1, content2)
    recordarray = ak.layout.RecordArray(
        {"one": content1, "two": listoffsetarray1, "three": content2}
    )

    offsets2 = ak.layout.Index64(np.array([0, 3, 3, 5], dtype=np.int64))
    listoffsetarray2 = ak.layout.ListOffsetArray64(offsets2, recordarray)

    listoffsetarray2 = v1_to_v2(listoffsetarray2)

    assert ak.to_list(listoffsetarray2["one"]) == [[1, 2, 3], [], [4, 5]]
    assert listoffsetarray2.typetracer["one"].form == listoffsetarray2["one"].form
    assert ak.to_list(listoffsetarray2["two"]) == [
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]],
        [],
        [[6.6], [7.7, 8.8, 9.9]],
    ]
    assert listoffsetarray2.typetracer["two"].form == listoffsetarray2["two"].form
    assert ak.to_list(listoffsetarray2["three"]) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert listoffsetarray2.typetracer["three"].form == listoffsetarray2["three"].form
    assert ak.to_list(listoffsetarray2[["two", "three"]]) == [
        [
            {"two": [1.1, 2.2, 3.3], "three": 1.1},
            {"two": [], "three": 2.2},
            {"two": [4.4, 5.5], "three": 3.3},
        ],
        [],
        [{"two": [6.6], "three": 4.4}, {"two": [7.7, 8.8, 9.9], "three": 5.5}],
    ]
    assert (
        listoffsetarray2.typetracer[["two", "three"]].form
        == listoffsetarray2[["two", "three"]].form
    )

    starts2 = ak.layout.Index64(np.array([0, 3, 3], dtype=np.int64))
    stops2 = ak.layout.Index64(np.array([3, 3, 5], dtype=np.int64))
    listarray2 = ak.layout.ListArray64(starts2, stops2, recordarray)
    listarray2 = v1_to_v2(listarray2)

    assert ak.to_list(listarray2["one"]) == [[1, 2, 3], [], [4, 5]]
    assert listarray2.typetracer["one"].form == listarray2["one"].form
    assert ak.to_list(listarray2["two"]) == [
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]],
        [],
        [[6.6], [7.7, 8.8, 9.9]],
    ]
    assert listarray2.typetracer["two"].form == listarray2["two"].form
    assert ak.to_list(listarray2["three"]) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert listarray2.typetracer["three"].form == listarray2["three"].form
    assert ak.to_list(listarray2[["two", "three"]]) == [
        [
            {"two": [1.1, 2.2, 3.3], "three": 1.1},
            {"two": [], "three": 2.2},
            {"two": [4.4, 5.5], "three": 3.3},
        ],
        [],
        [{"two": [6.6], "three": 4.4}, {"two": [7.7, 8.8, 9.9], "three": 5.5}],
    ]
    assert (
        listarray2.typetracer[["two", "three"]].form
        == listarray2[["two", "three"]].form
    )

    regulararray2 = ak.layout.RegularArray(recordarray, 1, zeros_length=0)
    regulararray2 = v1_to_v2(regulararray2)

    assert ak.to_list(regulararray2["one"]) == [[1], [2], [3], [4], [5]]
    assert regulararray2.typetracer["one"].form == regulararray2["one"].form
    assert ak.to_list(regulararray2["two"]) == [
        [[1.1, 2.2, 3.3]],
        [[]],
        [[4.4, 5.5]],
        [[6.6]],
        [[7.7, 8.8, 9.9]],
    ]
    assert regulararray2.typetracer["two"].form == regulararray2["two"].form
    assert ak.to_list(regulararray2["three"]) == [[1.1], [2.2], [3.3], [4.4], [5.5]]
    assert regulararray2.typetracer["three"].form == regulararray2["three"].form
    assert ak.to_list(regulararray2[["two", "three"]]) == [
        [{"two": [1.1, 2.2, 3.3], "three": 1.1}],
        [{"two": [], "three": 2.2}],
        [{"two": [4.4, 5.5], "three": 3.3}],
        [{"two": [6.6], "three": 4.4}],
        [{"two": [7.7, 8.8, 9.9], "three": 5.5}],
    ]
    assert (
        regulararray2.typetracer[["two", "three"]].form
        == regulararray2[["two", "three"]].form
    )


def test_getitem_next():
    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5], dtype=np.int64))
    content2 = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], dtype=np.float64)
    )
    content3 = ak.layout.NumpyArray(
        np.array([1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=np.float64)
    )
    offsets1 = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9], dtype=np.int64))
    listoffsetarray1 = ak.layout.ListOffsetArray64(offsets1, content2)
    listoffsetarray3 = ak.layout.ListOffsetArray64(offsets1, content3)
    recordarray = ak.layout.RecordArray(
        {
            "one": content1,
            "two": listoffsetarray1,
            "three": content2,
            "four": listoffsetarray3,
        }
    )
    offsets2 = ak.layout.Index64(np.array([0, 3, 3, 5], dtype=np.int64))
    listoffsetarray2 = ak.layout.ListOffsetArray64(offsets2, recordarray)

    listoffsetarray2 = v1_to_v2(listoffsetarray2)

    assert ak.to_list(listoffsetarray2[2, "one"]) == [4, 5]
    assert listoffsetarray2.typetracer[2, "one"].form == listoffsetarray2[2, "one"].form
    assert ak.to_list(listoffsetarray2[2, "two"]) == [[6.6], [7.7, 8.8, 9.9]]
    assert listoffsetarray2.typetracer[2, "two"].form == listoffsetarray2[2, "two"].form
    assert ak.to_list(listoffsetarray2[2, "three"]) == [4.4, 5.5]
    assert (
        listoffsetarray2.typetracer[2, "three"].form
        == listoffsetarray2[2, "three"].form
    )
    assert ak.to_list(listoffsetarray2[2, ["two", "three"]]) == [
        {"two": [6.6], "three": 4.4},
        {"two": [7.7, 8.8, 9.9], "three": 5.5},
    ]
    assert (
        listoffsetarray2.typetracer[2, ["two", "three"]].form
        == listoffsetarray2[2, ["two", "three"]].form
    )

    assert ak.to_list(listoffsetarray2[2, 1]) == {
        "one": 5,
        "two": [7.7, 8.8, 9.9],
        "three": 5.5,
        "four": [7, 8, 9],
    }
    assert (
        listoffsetarray2.typetracer[2, 1].array.form
        == listoffsetarray2[2, 1].array.form
    )
    with pytest.raises(IndexError):
        listoffsetarray2[2, 1, 0]
    assert listoffsetarray2[2, 1, "one"] == 5
    assert ak.to_list(listoffsetarray2[2, 1, "two"]) == [7.7, 8.8, 9.9]
    assert (
        listoffsetarray2.typetracer[2, 1, "two"].form
        == listoffsetarray2[2, 1, "two"].form
    )
    assert listoffsetarray2[2, 1, "two", 1] == 8.8
    assert ak.to_list(listoffsetarray2[2, 1, ["two", "four"], 1]) == {
        "two": 8.8,
        "four": 8,
    }
    assert (
        listoffsetarray2.typetracer[2, 1, ["two", "four"], 1].array.form
        == listoffsetarray2[2, 1, ["two", "four"], 1].array.form
    )
    assert ak.to_list(listoffsetarray2[2, 1, ["two", "four"], 1:]) == {
        "two": [8.8, 9.9],
        "four": [8, 9],
    }



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    np_data = np.random.random(size=(4, 100 * 1024 * 1024 // 8 // 4))
    array = ak.from_numpy(np_data, regulararray=False)
    assert np_data.nbytes == array.nbytes



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest
import numpy as np

import awkward as ak

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_index32():
    cpp_array = ak.layout.Index32(np.zeros(10))
    py_array = ak._v2.index.Index.zeros(10, np, np.int32)

    assert len(cpp_array) == len(py_array)
    assert "i32" == py_array.form


def test_index64():
    cpp_array = ak.layout.Index64(np.zeros(10))
    py_array = ak._v2.index.Index.zeros(10, np, np.int64)

    assert len(cpp_array) == len(py_array)
    assert "i64" == py_array.form


def test_identifier32():
    cpp_array = ak.layout.Identities32(123, [(1, "one"), (2, "two")], 10, 5)
    py_array = ak._v2.identifier.Identifier.zeros(
        123, {1: "one", 2: "two"}, 5, 10, np, np.int32
    )

    assert len(cpp_array) == len(py_array)
    assert py_array._data.dtype == np.dtype(np.int32)


def test_identifier64():
    cpp_array = ak.layout.Identities64(123, [(1, "one"), (2, "two")], 10, 5)
    py_array = ak._v2.identifier.Identifier.zeros(
        123, {1: "one", 2: "two"}, 5, 10, np, np.int64
    )

    assert len(cpp_array) == len(py_array)
    assert py_array._data.dtype == np.dtype(np.int64)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_numpy_array():
    matrix = np.arange(64).reshape(8, -1)
    layout = ak.layout.NumpyArray(matrix[:, 0])
    assert not layout.iscontiguous

    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)
    assert packed.iscontiguous


def test_empty_array():
    layout = ak.layout.EmptyArray()
    assert ak.packed(layout, highlevel=False) is layout


def test_indexed_option_array():
    index = ak.layout.Index64(np.r_[0, -1, 2, -1, 4])
    content = ak.layout.NumpyArray(np.arange(8))
    layout = ak.layout.IndexedOptionArray64(index, content)

    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(layout) == ak.to_list(packed)
    assert isinstance(packed, ak.layout.IndexedOptionArray64)
    assert np.asarray(packed.index).tolist() == [0, -1, 1, -1, 2]
    assert len(packed.content) == 3


def test_indexed_array():
    index = ak.layout.Index64(np.array([0, 1, 2, 3, 6, 7, 8]))
    content = ak.layout.NumpyArray(np.arange(10))
    layout = ak.layout.IndexedArray64(index, content)

    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)

    assert isinstance(packed, ak.layout.NumpyArray)
    assert len(packed) == len(index)


def test_list_array():
    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts = ak.layout.Index64(np.array([0, 3, 3, 5, 6]))
    stops = ak.layout.Index64(np.array([3, 3, 5, 6, 9]))
    layout = ak.layout.ListArray64(starts, stops, content)

    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)
    assert isinstance(packed, ak.layout.ListOffsetArray64)
    assert packed.offsets[0] == 0


def test_list_offset_array():
    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6]))
    layout = ak.layout.ListOffsetArray64(offsets, content)

    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)
    assert isinstance(packed, ak.layout.ListOffsetArray64)
    assert packed.offsets[0] == 0
    assert len(packed.content) == packed.offsets[-1]


def test_unmasked_array():
    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    layout = ak.layout.UnmaskedArray(content)
    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)


def test_union_array():
    a = ak.layout.NumpyArray(np.arange(4))
    b = ak.layout.NumpyArray(np.arange(4) + 4)
    c = ak.layout.RegularArray(ak.layout.NumpyArray(np.arange(12)), 3)
    layout = ak.layout.UnionArray8_64(
        ak.layout.Index8([1, 1, 2, 2, 0, 0]),
        ak.layout.Index64([0, 1, 0, 1, 0, 1]),
        [a, b, c],
    )
    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)
    # Check that it merges like contents
    assert len(packed.contents) == 2
    index_0 = np.asarray(packed.index)[np.asarray(packed.tags) == 0]
    assert index_0.tolist() == [0, 1, 2, 3]


def test_record_array():
    a = ak.layout.NumpyArray(np.arange(10))
    b = ak.layout.NumpyArray(np.arange(10) * 2 + 4)
    layout = ak.layout.RecordArray([a, b], None, 5)
    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)
    assert len(packed.contents[0]) == 5
    assert len(packed.contents[1]) == 5


def test_regular_array():
    content = ak.layout.NumpyArray(np.arange(10))
    layout = ak.layout.RegularArray(content, 3)
    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)
    assert len(packed.content) == 9
    assert packed.size == layout.size


def test_bit_masked_aray():
    mask = ak.layout.IndexU8(np.array([0b10101010]))
    content = ak.layout.NumpyArray(np.arange(16))
    layout = ak.layout.BitMaskedArray(mask, content, False, 8, False)
    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)
    assert len(packed.content) == 8


def test_byte_masked_array():
    mask = ak.layout.Index8(np.array([1, 0, 1, 0, 1, 0, 1, 0]))
    content = ak.layout.NumpyArray(np.arange(16))
    layout = ak.layout.ByteMaskedArray(
        mask,
        content,
        False,
    )
    packed = ak.packed(layout, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(layout)
    assert len(packed.content) == 8


def test_virtual_array():
    n_called = [0]

    def generate():
        n_called[0] += 1
        return ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))

    generator = ak.layout.ArrayGenerator(
        generate, form=ak.forms.NumpyForm([], 8, "d"), length=5
    )
    layout = ak.layout.VirtualArray(generator)
    assert n_called[0] == 0
    packed = ak.packed(layout, highlevel=False)
    assert n_called[0] == 1

    assert isinstance(packed, ak.layout.NumpyArray)
    assert ak.to_list(packed) == ak.to_list(layout)


def test_partitioned_array():
    index_0 = ak.layout.Index64(np.array([0, 1, 2, 3, 6, 7, 8]))
    content_0 = ak.layout.NumpyArray(np.arange(10))
    content = ak.layout.IndexedArray64(index_0, content_0)
    layout = ak.partitioned([content, content], highlevel=False)
    packed = ak.packed(layout, highlevel=False)

    assert ak.to_list(layout) == ak.to_list(packed)
    assert isinstance(packed, ak.partition.PartitionedArray)

    assert isinstance(packed.partitions[0], ak.layout.NumpyArray)
    assert len(packed.partitions[0]) == len(index_0)

    assert isinstance(packed.partitions[1], ak.layout.NumpyArray)
    assert len(packed.partitions[1]) == len(index_0)


def test_record():
    a = ak.layout.NumpyArray(np.arange(10))
    b = ak.layout.NumpyArray(np.arange(10) * 2 + 4)
    layout = ak.layout.RecordArray([a, b], None, 5)
    record = layout[4]
    packed = ak.packed(record, highlevel=False)
    assert ak.to_list(packed) == ak.to_list(record)
    assert len(packed.array) == 1
    assert ak.to_list(packed.array) == [(4, 12)]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

from awkward._ext import ForthMachine32
from awkward._ext import ForthMachine64

__all__ = ["ForthMachine32", "ForthMachine64"]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    arr1 = ak.Array({"a": [1, 2], "b": [1, None]})
    arr2 = ak.mask(arr1, [True, True])
    assert isinstance(arr2.layout, ak.layout.ByteMaskedArray)
    assert isinstance(arr2.layout.content, ak.layout.RecordArray)
    assert isinstance(arr2.layout.content.field("b"), ak.layout.IndexedOptionArray64)

    assert isinstance(arr2.b.layout, ak.layout.IndexedOptionArray64)
    assert isinstance(arr2.b.layout.content, ak.layout.NumpyArray)

    assert ak.is_none(arr2.b).tolist() == [False, True]

    arr3 = ak.virtual(lambda: arr2, form=arr2.layout.form, length=len(arr2))
    assert ak.is_none(arr3.b).tolist() == [False, True]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import sys
import itertools

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_refcount():
    o = np.arange(10, dtype="i4")
    c = np.arange(12).reshape(3, 4)

    for order in itertools.permutations(["del i, n", "del l1", "del l2"]):
        i = ak.layout.Index32(o)
        n = ak.layout.NumpyArray(c)
        l1 = ak.layout.ListOffsetArray32(i, n)
        l2 = ak.layout.ListOffsetArray32(i, l1)  # noqa: F841 (checking refcount)

        for statement in order:
            assert sys.getrefcount(o), sys.getrefcount(c) == (3, 3)
            exec(statement)
            assert sys.getrefcount(o), sys.getrefcount(c) == (2, 2)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_flatten():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10], dtype=np.int64))
    array = ak.layout.ListOffsetArray64(offsets, content)

    assert ak.to_list(array) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert ak.to_list(array.flatten(axis=1)) == [
        0.0,
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        7.7,
        8.8,
        9.9,
    ]
    assert ak.to_list(array.flatten(axis=-1)) == [
        0.0,
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        7.7,
        8.8,
        9.9,
    ]
    with pytest.raises(ValueError) as err:
        assert ak.to_list(array.flatten(axis=-2))
    assert str(err.value).startswith("axis=0 not allowed for flatten")

    array2 = array[2:-1]
    assert ak.to_list(array2.flatten(axis=1)) == [3.3, 4.4, 5.5]
    assert ak.to_list(array2.flatten(axis=-1)) == [3.3, 4.4, 5.5]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

numba = pytest.importorskip("numba")


def test():
    @numba.njit
    def f1(array):
        return array.ndim

    assert f1(ak.Array([[1, 2, 3], [], [4, 5]])) == 2
    assert f1(ak.Array([[[1], [2, 3]], [], [[4, 5], []]])) == 3

    with pytest.raises(numba.core.errors.TypingError):
        f1(ak.Record({"x": [1, 2, 3], "y": [4]}))

    partitioned = ak.partitioned([ak.Array([[1, 2, 3], [], [4, 5]]) for i in range(5)])

    assert f1(partitioned) == 2



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import cupy as cp  # noqa: F401
import awkward as ak  # noqa: F401


def test_tocuda():
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    bitmask = ak.layout.IndexU8(np.array([40, 34], dtype=np.uint8))
    array = ak.layout.BitMaskedArray(bitmask, content, False, 9, False)
    cuda_array = ak.to_kernels(array, "cuda")
    copyback_array = ak.to_kernels(cuda_array, "cpu")
    assert ak.to_list(cuda_array) == ak.to_list(array)
    assert ak.to_list(copyback_array) == ak.to_list(array)

    bytemask = ak.layout.Index8(np.array([False, True, False], dtype=np.bool))
    array = ak.layout.ByteMaskedArray(bytemask, content, True)
    cuda_array = ak.to_kernels(array, "cuda")
    copyback_array = ak.to_kernels(cuda_array, "cpu")
    assert ak.to_list(cuda_array) == ak.to_list(array)
    assert ak.to_list(copyback_array) == ak.to_list(array)

    array = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5]))
    cuda_array = ak.to_kernels(array, "cuda")
    copyback_array = ak.to_kernels(cuda_array, "cpu")
    assert ak.to_list(cuda_array) == ak.to_list(array)
    assert ak.to_list(copyback_array) == ak.to_list(array)

    array = ak.layout.NumpyArray(np.array([[0.0, 1.1], [2.2, 3.3], [4.4, 5.5]]))
    cuda_array = ak.to_kernels(array, "cuda")
    copyback_array = ak.to_kernels(cuda_array, "cpu")

    assert ak.to_list(cuda_array) == ak.to_list(array)
    assert ak.to_list(copyback_array) == ak.to_list(array)

    array = ak.layout.EmptyArray()
    cuda_array = ak.to_kernels(array, "cuda")
    copyback_array = ak.to_kernels(cuda_array, "cpu")
    assert ak.to_list(cuda_array) == ak.to_list(array)
    assert ak.to_list(copyback_array) == ak.to_list(array)

    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9]))
    array = ak.layout.ListOffsetArray64(offsets, content)
    cuda_array = ak.to_kernels(array, "cuda")
    copyback_array = ak.to_kernels(cuda_array, "cpu")
    assert ak.to_list(cuda_array) == ak.to_list(array)
    assert ak.to_list(copyback_array) == ak.to_list(array)

    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.IndexU32(np.array([0, 3, 3, 5, 6, 9]))
    array = ak.layout.ListOffsetArrayU32(offsets, content)
    cuda_array = ak.to_kernels(array, "cuda")
    copyback_array = ak.to_kernels(cuda_array, "cpu")
    assert ak.to_list(cuda_array) == ak.to_list(array)
    assert ak.to_list(copyback_array) == ak.to_list(array)

    # # Testing parameters
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    offsets = ak.layout.Index32(np.array([0, 3, 3, 5, 6, 9]))
    array = ak.layout.ListOffsetArray32(offsets, content)
    cuda_array = ak.to_kernels(array, "cuda")
    copyback_array = ak.to_kernels(cuda_array, "cpu")
    assert ak.to_list(cuda_array) == ak.to_list(array)
    assert ak.to_list(copyback_array) == ak.to_list(array)

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    regulararray = ak.layout.RegularArray(listoffsetarray, 2)
    starts = ak.layout.Index64(np.array([0, 1]))
    stops = ak.layout.Index64(np.array([2, 3]))
    listarray = ak.layout.ListArray64(starts, stops, regulararray)

    cuda_listoffsetarray = ak.to_kernels(listoffsetarray, "cuda")
    copyback_listoffsetarray = ak.to_kernels(cuda_listoffsetarray, "cpu")
    assert ak.to_list(cuda_listoffsetarray) == ak.to_list(listoffsetarray)
    assert ak.to_list(copyback_listoffsetarray) == ak.to_list(listoffsetarray)

    cuda_regulararray = ak.to_kernels(regulararray, "cuda")
    copyback_regulararray = ak.to_kernels(cuda_regulararray, "cpu")
    assert ak.to_list(cuda_regulararray) == ak.to_list(regulararray)
    assert ak.to_list(copyback_regulararray) == ak.to_list(regulararray)

    cuda_listarray = ak.to_kernels(listarray, "cuda")
    copyback_listarray = ak.to_kernels(cuda_listarray, "cpu")
    assert ak.to_list(cuda_listarray) == ak.to_list(listarray)
    assert ak.to_list(copyback_listarray) == ak.to_list(listarray)

    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5]))
    content2 = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.Index32(np.array([0, 3, 3, 5, 6, 9]))

    recordarray = ak.layout.RecordArray(
        [content1, listoffsetarray, content2, content1],
        keys=["one", "two", "2", "wonky"],
    )

    cuda_recordarray = ak.to_kernels(recordarray, "cuda")
    copyback_recordarray = ak.to_kernels(cuda_recordarray, "cpu")
    assert ak.to_list(cuda_recordarray) == ak.to_list(recordarray)
    assert ak.to_list(copyback_recordarray) == ak.to_list(recordarray)

    content0 = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]]).layout
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    tags = ak.layout.Index8(np.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=np.int8))
    index = ak.layout.Index32(np.array([0, 1, 0, 1, 2, 2, 4, 3], dtype=np.int32))
    unionarray = ak.layout.UnionArray8_32(tags, index, [content0, content1])

    cuda_unionarray = ak.to_kernels(unionarray, "cuda")
    copyback_unionarray = ak.to_kernels(cuda_unionarray, "cpu")
    assert ak.to_list(cuda_unionarray) == ak.to_list(unionarray)
    assert ak.to_list(copyback_unionarray) == ak.to_list(unionarray)

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    index = ak.layout.Index32(np.array([0, 2, 4, 6, 8, 9, 7, 5], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray32(index, content)

    cuda_indexedarray = ak.to_kernels(indexedarray, "cuda")
    copyback_indexedarray = ak.to_kernels(cuda_indexedarray, "cpu")
    assert ak.to_list(cuda_indexedarray) == ak.to_list(indexedarray)
    assert ak.to_list(copyback_indexedarray) == ak.to_list(indexedarray)

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, True, False, False, False], dtype=np.int8)),
        listoffsetarray,
        True,
    )
    cuda_bytemaskedarray = ak.to_kernels(bytemaskedarray, "cuda")
    copyback_bytemaskedarray = ak.to_kernels(cuda_bytemaskedarray, "cpu")
    assert ak.to_list(cuda_bytemaskedarray) == ak.to_list(bytemaskedarray)
    assert ak.to_list(copyback_bytemaskedarray) == ak.to_list(bytemaskedarray)

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, False], dtype=np.int8)), listarray, True
    )
    cuda_bytemaskedarray = ak.to_kernels(bytemaskedarray, "cuda")
    copyback_bytemaskedarray = ak.to_kernels(cuda_bytemaskedarray, "cpu")
    assert ak.to_list(cuda_bytemaskedarray) == ak.to_list(bytemaskedarray)
    assert ak.to_list(copyback_bytemaskedarray) == ak.to_list(bytemaskedarray)

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, False], dtype=np.int8)), recordarray, True
    )
    cuda_bytemaskedarray = ak.to_kernels(bytemaskedarray, "cuda")
    copyback_bytemaskedarray = ak.to_kernels(cuda_bytemaskedarray, "cpu")
    assert ak.to_list(cuda_bytemaskedarray) == ak.to_list(bytemaskedarray)
    assert ak.to_list(copyback_bytemaskedarray) == ak.to_list(bytemaskedarray)

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, False, False], dtype=np.int8)),
        indexedarray,
        True,
    )
    cuda_bytemaskedarray = ak.to_kernels(bytemaskedarray, "cuda")
    copyback_bytemaskedarray = ak.to_kernels(cuda_bytemaskedarray, "cpu")
    assert ak.to_list(cuda_bytemaskedarray) == ak.to_list(bytemaskedarray)
    assert ak.to_list(copyback_bytemaskedarray) == ak.to_list(bytemaskedarray)

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, False, False], dtype=np.int8)),
        unionarray,
        True,
    )
    cuda_bytemaskedarray = ak.to_kernels(bytemaskedarray, "cuda")
    copyback_bytemaskedarray = ak.to_kernels(cuda_bytemaskedarray, "cpu")
    assert ak.to_list(cuda_bytemaskedarray) == ak.to_list(bytemaskedarray)
    assert ak.to_list(copyback_bytemaskedarray) == ak.to_list(bytemaskedarray)

    ioa = ak.layout.IndexedOptionArray32(
        ak.layout.Index32([-30, 19, 6, 7, -3, 21, 13, 22, 17, 9, -12, 16]),
        ak.layout.NumpyArray(
            np.array(
                [
                    5.2,
                    1.7,
                    6.7,
                    -0.4,
                    4.0,
                    7.8,
                    3.8,
                    6.8,
                    4.2,
                    0.3,
                    4.6,
                    6.2,
                    6.9,
                    -0.7,
                    3.9,
                    1.6,
                    8.7,
                    -0.7,
                    3.2,
                    4.3,
                    4.0,
                    5.8,
                    4.2,
                    7.0,
                    5.6,
                    3.8,
                ]
            )
        ),
    )
    cuda_ioa = ak.to_kernels(ioa, "cuda")
    copyback_ioa = ak.to_kernels(cuda_ioa, "cpu")
    assert ak.to_list(cuda_ioa) == ak.to_list(ioa)
    assert ak.to_list(copyback_ioa) == ak.to_list(ioa)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array(
        [
            [0, 1, None],
            [None, 3],
            [],
            None,
            [4, 5, None, 6],
        ]
    )

    assert ak.sum(array, axis=-1).tolist() == [
        1,  # 0 + 1
        3,  # 3
        0,  # list is empty
        None,  # list is missing
        15,  # 4 + 5 + 6
    ]

    assert ak.sum(array, axis=-2).tolist() == [
        4,
        9,
        0,
        6,
        # 0+4
        #     1+3+5
        #         no data
        #                     6
    ]

    assert ak.min(array, axis=-1).tolist() == [
        0,  # min([0, 1])
        3,  # min([3])
        None,  # list is empty
        None,  # list is missing
        4,  # min([4, 5, 6])
    ]

    assert ak.min(array, axis=-2).tolist() == [
        0,
        1,
        None,
        6,
        # 0,4
        #     1,3,5
        #          no data
        #                     6
    ]

    # first bug-fix: single '?'
    assert str(ak.min(array, axis=-1).type) == "5 * ?int64"

    # second bug-fix: correct mask_identity=False behavior
    assert ak.ptp(array, axis=-1).tolist() == [1, 0, None, None, 2]
    assert ak.ptp(array, axis=-1, mask_identity=False).tolist() == [1, 0, 0, None, 2]

    assert ak.ptp(array, axis=-2).tolist() == [4, 4, None, 0]
    assert ak.ptp(array, axis=-2, mask_identity=False).tolist() == [4, 4, 0, 0]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

try:
    # pybind11 only supports cPickle protocol 2+ (-1 in pickle.dumps)
    # (automatically satisfied in Python 3; this is just to keep testing Python 2.7)
    import cPickle as pickle
except ImportError:
    import pickle

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_numpyarray():
    assert ak.from_buffers(*ak.to_buffers([1, 2, 3, 4, 5])).tolist() == [
        1,
        2,
        3,
        4,
        5,
    ]
    assert pickle.loads(pickle.dumps(ak.Array([1, 2, 3, 4, 5]), -1)).tolist() == [
        1,
        2,
        3,
        4,
        5,
    ]


def test_listoffsetarray():
    assert ak.from_buffers(*ak.to_buffers([[1, 2, 3], [], [4, 5]])).tolist() == [
        [1, 2, 3],
        [],
        [4, 5],
    ]
    assert ak.from_buffers(
        *ak.to_buffers(["one", "two", "three", "four", "five"])
    ).tolist() == ["one", "two", "three", "four", "five"]
    assert ak.from_buffers(
        *ak.to_buffers([["one", "two", "three"], [], ["four", "five"]])
    ).tolist() == [["one", "two", "three"], [], ["four", "five"]]
    assert pickle.loads(
        pickle.dumps(ak.Array([[1, 2, 3], [], [4, 5]]), -1)
    ).tolist() == [[1, 2, 3], [], [4, 5]]


def test_listarray():
    listoffsetarray = ak.Array([[1, 2, 3], [], [4, 5]]).layout
    listarray = ak.layout.ListArray64(
        listoffsetarray.starts, listoffsetarray.stops, listoffsetarray.content
    )
    assert ak.from_buffers(*ak.to_buffers(listarray)).tolist() == [
        [1, 2, 3],
        [],
        [4, 5],
    ]
    assert pickle.loads(pickle.dumps(ak.Array(listarray), -1)).tolist() == [
        [1, 2, 3],
        [],
        [4, 5],
    ]


def test_indexedoptionarray():
    assert ak.from_buffers(*ak.to_buffers([1, 2, 3, None, None, 5])).tolist() == [
        1,
        2,
        3,
        None,
        None,
        5,
    ]
    assert pickle.loads(
        pickle.dumps(ak.Array([1, 2, 3, None, None, 5]), -1)
    ).tolist() == [1, 2, 3, None, None, 5]


def test_indexedarray():
    content = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4]).layout
    index = ak.layout.Index64(np.array([3, 1, 1, 4, 2], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray64(index, content)
    assert ak.from_buffers(*ak.to_buffers(indexedarray)).tolist() == [
        3.3,
        1.1,
        1.1,
        4.4,
        2.2,
    ]
    assert pickle.loads(pickle.dumps(ak.Array(indexedarray), -1)).tolist() == [
        3.3,
        1.1,
        1.1,
        4.4,
        2.2,
    ]


def test_emptyarray():
    assert ak.from_buffers(*ak.to_buffers([])).tolist() == []
    assert ak.from_buffers(*ak.to_buffers([[], [], []])).tolist() == [[], [], []]

    assert pickle.loads(pickle.dumps(ak.Array([]), -1)).tolist() == []
    assert pickle.loads(pickle.dumps(ak.Array([[], [], []]), -1)).tolist() == [
        [],
        [],
        [],
    ]


def test_bytemaskedarray():
    content = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4]).layout
    mask = ak.layout.Index8(np.array([False, True, True, False, False], dtype=np.int8))
    bytemaskedarray = ak.layout.ByteMaskedArray(mask, content, True)
    assert ak.from_buffers(*ak.to_buffers(bytemaskedarray)).tolist() == [
        None,
        1.1,
        2.2,
        None,
        None,
    ]
    assert pickle.loads(pickle.dumps(ak.Array(bytemaskedarray), -1)).tolist() == [
        None,
        1.1,
        2.2,
        None,
        None,
    ]


def test_bitmaskedarray():
    content = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4]).layout
    mask = ak.layout.IndexU8(
        np.packbits(np.array([False, True, True, False, False], dtype=np.int8))
    )
    bitmaskedarray = ak.layout.BitMaskedArray(mask, content, True, 5, False)
    assert ak.from_buffers(*ak.to_buffers(bitmaskedarray)).tolist() == [
        None,
        1.1,
        2.2,
        None,
        None,
    ]
    assert pickle.loads(pickle.dumps(ak.Array(bitmaskedarray), -1)).tolist() == [
        None,
        1.1,
        2.2,
        None,
        None,
    ]


def test_recordarray():
    assert ak.from_buffers(
        *ak.to_buffers([(1.1, [1]), (2.2, [1, 2]), (3.3, [1, 2, 3])])
    ).tolist() == [(1.1, [1]), (2.2, [1, 2]), (3.3, [1, 2, 3])]
    assert ak.from_buffers(
        *ak.to_buffers(
            [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}, {"x": 3.3, "y": [1, 2, 3]}]
        )
    ).tolist() == [
        {"x": 1.1, "y": [1]},
        {"x": 2.2, "y": [1, 2]},
        {"x": 3.3, "y": [1, 2, 3]},
    ]

    assert pickle.loads(
        pickle.dumps(ak.Array([(1.1, [1]), (2.2, [1, 2]), (3.3, [1, 2, 3])]), -1)
    ).tolist() == [(1.1, [1]), (2.2, [1, 2]), (3.3, [1, 2, 3])]
    assert pickle.loads(
        pickle.dumps(
            ak.Array(
                [
                    {"x": 1.1, "y": [1]},
                    {"x": 2.2, "y": [1, 2]},
                    {"x": 3.3, "y": [1, 2, 3]},
                ]
            ),
            -1,
        )
    ).tolist() == [
        {"x": 1.1, "y": [1]},
        {"x": 2.2, "y": [1, 2]},
        {"x": 3.3, "y": [1, 2, 3]},
    ]


def test_record():
    assert pickle.loads(
        pickle.dumps(ak.Record({"x": 2.2, "y": [1, 2]}), -1)
    ).tolist() == {"x": 2.2, "y": [1, 2]}
    assert (
        pickle.loads(
            pickle.dumps(
                ak.Array(
                    [
                        {"x": 1.1, "y": [1]},
                        {"x": 2.2, "y": [1, 2]},
                        {"x": 3.3, "y": [1, 2, 3]},
                    ]
                )[1],
                -1,
            )
        ).tolist()
        == {"x": 2.2, "y": [1, 2]}
    )


def test_regulararray():
    content = ak.Array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]).layout
    regulararray = ak.layout.RegularArray(content, 3, zeros_length=0)
    assert ak.from_buffers(*ak.to_buffers(regulararray)).tolist() == [
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
        [10, 11, 12],
    ]
    assert pickle.loads(pickle.dumps(ak.Array(regulararray), -1)).tolist() == [
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
        [10, 11, 12],
    ]


def test_unionarray():
    assert ak.from_buffers(*ak.to_buffers([[1, 2, 3], [], 4, 5])).tolist() == [
        [1, 2, 3],
        [],
        4,
        5,
    ]
    assert pickle.loads(pickle.dumps(ak.Array([[1, 2, 3], [], 4, 5]), -1)).tolist() == [
        [1, 2, 3],
        [],
        4,
        5,
    ]


def test_unmaskedarray():
    content = ak.Array([1, 2, 3, 4, 5]).layout
    unmaskedarray = ak.layout.UnmaskedArray(content)
    assert ak.from_buffers(*ak.to_buffers(unmaskedarray)).tolist() == [1, 2, 3, 4, 5]
    assert pickle.loads(pickle.dumps(ak.Array(unmaskedarray), -1)).tolist() == [
        1,
        2,
        3,
        4,
        5,
    ]


def test_partitioned():
    array = ak.repartition(ak.Array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]), 3)

    form, length, container = ak.to_buffers(array)
    assert ak.from_buffers(form, length, container).tolist() == [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
    ]

    form, length, container = ak.to_buffers(array)
    assert ak.from_buffers(form, length, container).tolist() == [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
    ]

    one = ak.Array([1, 2, 3, 4, 5])
    two = ak.Array([6, 7, 8, 9, 10])
    container = {}
    lengths = []
    form1, length, _ = ak.to_buffers(one, container, 0)
    lengths.append(length)
    form2, length, _ = ak.to_buffers(two, container, 1)
    lengths.append(length)
    assert form1 == form2

    assert ak.from_buffers(form1, lengths, container).tolist() == [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
    ]
    assert pickle.loads(pickle.dumps(array, -1)).tolist() == [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
    ]


def test_lazy():
    array = ak.Array([1, 2, 3, 4, 5])

    form, length, container = ak.to_buffers(array)

    assert ak.from_buffers(form, length, container, lazy=True).tolist() == [
        1,
        2,
        3,
        4,
        5,
    ]


def test_lazy_partitioned():
    array = ak.repartition(ak.Array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]), 3)
    form, length, container = ak.to_buffers(array)
    assert length == [3, 3, 3, 1]

    assert ak.from_buffers(form, length, container, lazy=True).tolist() == [
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

from awkward._ext import Form
from awkward._ext import BitMaskedForm
from awkward._ext import ByteMaskedForm
from awkward._ext import EmptyForm
from awkward._ext import IndexedForm
from awkward._ext import IndexedOptionForm
from awkward._ext import ListForm
from awkward._ext import ListOffsetForm
from awkward._ext import NumpyForm
from awkward._ext import RecordForm
from awkward._ext import RegularForm
from awkward._ext import UnionForm
from awkward._ext import UnmaskedForm
from awkward._ext import VirtualForm


__all__ = [
    "Form",
    "BitMaskedForm",
    "ByteMaskedForm",
    "EmptyForm",
    "IndexedForm",
    "IndexedOptionForm",
    "ListForm",
    "ListOffsetForm",
    "NumpyForm",
    "RecordForm",
    "RegularForm",
    "UnionForm",
    "UnmaskedForm",
    "VirtualForm",
]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
from awkward._v2.index import Index
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.contents.listoffsetarray import ListOffsetArray
from awkward._v2.forms.listform import ListForm

np = ak.nplike.NumpyMetadata.instance()


class ListArray(Content):
    def __init__(self, starts, stops, content, identifier=None, parameters=None):
        if not isinstance(starts, Index) and starts.dtype in (
            np.dtype(np.int32),
            np.dtype(np.uint32),
            np.dtype(np.int64),
        ):
            raise TypeError(
                "{0} 'starts' must be an Index with dtype in (int32, uint32, int64), "
                "not {1}".format(type(self).__name__, repr(starts))
            )
        if not (isinstance(stops, Index) and starts.dtype == stops.dtype):
            raise TypeError(
                "{0} 'stops' must be an Index with the same dtype as 'starts' ({1}), "
                "not {2}".format(type(self).__name__, repr(starts.dtype), repr(stops))
            )
        if not isinstance(content, Content):
            raise TypeError(
                "{0} 'content' must be a Content subtype, not {1}".format(
                    type(self).__name__, repr(content)
                )
            )
        if not len(starts) <= len(stops):
            raise ValueError(
                "{0} len(starts) ({1}) must be <= len(stops) ({2})".format(
                    type(self).__name__, len(starts), len(stops)
                )
            )

        self._starts = starts
        self._stops = stops
        self._content = content
        self._init(identifier, parameters)

    @property
    def starts(self):
        return self._starts

    @property
    def stops(self):
        return self._stops

    @property
    def content(self):
        return self._content

    @property
    def nplike(self):
        return self._starts.nplike

    @property
    def nonvirtual_nplike(self):
        return self._starts.nplike

    Form = ListForm

    @property
    def form(self):
        return self.Form(
            self._starts.form,
            self._stops.form,
            self._content.form,
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return ListArray(
            ak._v2.index.Index(self._starts.to(tt)),
            ak._v2.index.Index(self._stops.to(tt)),
            self._content.typetracer,
            self._typetracer_identifier(),
            self._parameters,
        )

    def __len__(self):
        return len(self._starts)

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<ListArray len="]
        out.append(repr(str(len(self))))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._starts._repr(indent + "    ", "<starts>", "</starts>\n"))
        out.append(self._stops._repr(indent + "    ", "<stops>", "</stops>\n"))
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</ListArray>")
        out.append(post)
        return "".join(out)

    def toListOffsetArray64(self, start_at_zero=False):
        offsets = self._compact_offsets64(start_at_zero)
        return self._broadcast_tooffsets64(offsets)

    def toRegularArray(self):
        offsets = self._compact_offsets64(True)
        return self._broadcast_tooffsets64(offsets).toRegularArray()

    def _getitem_nothing(self):
        return self._content._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if where < 0:
            where += len(self)
        if not (0 <= where < len(self)) and self.nplike.known_shape:
            raise NestedIndexError(self, where)
        start, stop = self._starts[where], self._stops[where]
        return self._content._getitem_range(slice(start, stop))

    def _getitem_range(self, where):
        start, stop, step = where.indices(len(self))
        assert step == 1
        return ListArray(
            self._starts[start:stop],
            self._stops[start:stop],
            self._content,
            self._range_identifier(start, stop),
            self._parameters,
        )

    def _getitem_field(self, where, only_fields=()):
        return ListArray(
            self._starts,
            self._stops,
            self._content._getitem_field(where, only_fields),
            self._field_identifier(where),
            None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return ListArray(
            self._starts,
            self._stops,
            self._content._getitem_fields(where, only_fields),
            self._fields_identifier(where),
            None,
        )

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        try:
            nextstarts = self._starts[carry.data]
            nextstops = self._stops[: len(self._starts)][carry.data]
        except IndexError as err:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, str(err))
            else:
                raise exception(str(err))

        return ListArray(
            nextstarts,
            nextstops,
            self._content,
            self._carry_identifier(carry, exception),
            self._parameters,
        )

    def _compact_offsets64(self, start_at_zero):
        starts_len = len(self._starts)
        out = ak._v2.index.Index64.empty(starts_len + 1, self.nplike)
        self._handle_error(
            self.nplike[
                "awkward_ListArray_compact_offsets",
                out.dtype.type,
                self._starts.dtype.type,
                self._stops.dtype.type,
            ](
                out.to(self.nplike),
                self._starts.to(self.nplike),
                self._stops.to(self.nplike),
                starts_len,
            )
        )
        if isinstance(out.data, ak._v2._typetracer.TypeTracerArray):
            out.data.fill_other = len(self._content)
        return out

    def _broadcast_tooffsets64(self, offsets):
        return ListOffsetArray._broadcast_tooffsets64(self, offsets)

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        nplike = self.nplike
        if len(slicestarts) != len(self) and nplike.known_shape:
            raise NestedIndexError(
                self,
                ak._v2.contents.ListArray(slicestarts, slicestops, slicecontent),
                "cannot fit jagged slice with length {0} into {1} of size {2}".format(
                    len(slicestarts), type(self).__name__, len(self)
                ),
            )

        if isinstance(slicecontent, ak._v2.contents.listoffsetarray.ListOffsetArray):
            outoffsets = ak._v2.index.Index64.empty(len(slicestarts) + 1, nplike)
            self._handle_error(
                nplike[
                    "awkward_ListArray_getitem_jagged_descend",
                    outoffsets.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    outoffsets.to(nplike),
                    slicestarts.to(nplike),
                    slicestops.to(nplike),
                    len(slicestarts),
                    self._starts.to(nplike),
                    self._stops.to(nplike),
                )
            )

            asListOffsetArray64 = self.toListOffsetArray64(True)
            next_content = asListOffsetArray64._content

            sliceoffsets = ak._v2.index.Index64(slicecontent._offsets)

            outcontent = next_content._getitem_next_jagged(
                sliceoffsets[:-1], sliceoffsets[1:], slicecontent._content, tail
            )

            return ak._v2.contents.listoffsetarray.ListOffsetArray(
                outoffsets, outcontent, None, self._parameters
            )

        elif isinstance(slicecontent, ak._v2.contents.numpyarray.NumpyArray):
            carrylen = ak._v2.index.Index64.empty(1, nplike)
            self._handle_error(
                nplike[
                    "awkward_ListArray_getitem_jagged_carrylen",
                    carrylen.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                ](
                    carrylen.to(nplike),
                    slicestarts.to(nplike),
                    slicestops.to(nplike),
                    len(slicestarts),
                )
            )
            sliceindex = ak._v2.index.Index64(slicecontent._data)
            outoffsets = ak._v2.index.Index64.zeros(len(slicestarts) + 1, nplike)
            nextcarry = ak._v2.index.Index64.zeros(carrylen[0], nplike)

            self._handle_error(
                nplike[
                    "awkward_ListArray_getitem_jagged_apply",
                    outoffsets.dtype.type,
                    nextcarry.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                    sliceindex.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    outoffsets.to(nplike),
                    nextcarry.to(nplike),
                    slicestarts.to(nplike),
                    slicestops.to(nplike),
                    len(slicestarts),
                    sliceindex.to(nplike),
                    len(sliceindex),
                    self._starts.to(nplike),
                    self._stops.to(nplike),
                    len(self._content),
                )
            )

            nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
            nexthead, nexttail = ak._v2._slicing.headtail(tail)
            outcontent = nextcontent._getitem_next(nexthead, nexttail, None)

            return ak._v2.contents.listoffsetarray.ListOffsetArray(
                outoffsets, outcontent
            )

        elif isinstance(
            slicecontent, ak._v2.contents.indexedoptionarray.IndexedOptionArray
        ):
            if len(self._starts) < len(slicestarts) and nplike.known_shape:
                raise NestedIndexError(
                    self,
                    ak._v2.contents.ListArray(slicestarts, slicestops, slicecontent),
                    "jagged slice length differs from array length",
                )

            missing = ak._v2.index.Index64(slicecontent._index)
            numvalid = ak._v2.index.Index64.empty(1, nplike)
            self._handle_error(
                nplike[
                    "awkward_ListArray_getitem_jagged_numvalid",
                    numvalid.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                    missing.dtype.type,
                ](
                    numvalid.to(nplike),
                    slicestarts.to(nplike),
                    slicestops.to(nplike),
                    len(slicestarts),
                    missing.to(nplike),
                    len(missing),
                )
            )

            nextcarry = ak._v2.index.Index64.empty(numvalid[0], nplike)
            if isinstance(nextcarry.data, ak._v2._typetracer.TypeTracerArray):
                nextcarry.data.shape = ak._v2._typetracer.Interval(
                    0, len(slicecontent._index)
                )

            smalloffsets = ak._v2.index.Index64.empty(len(slicestarts) + 1, nplike)
            largeoffsets = ak._v2.index.Index64.empty(len(slicestarts) + 1, nplike)

            self._handle_error(
                nplike[
                    "awkward_ListArray_getitem_jagged_shrink",
                    nextcarry.dtype.type,
                    smalloffsets.dtype.type,
                    largeoffsets.dtype.type,
                    slicestarts.dtype.type,
                    slicestops.dtype.type,
                    missing.dtype.type,
                ](
                    nextcarry.to(nplike),
                    smalloffsets.to(nplike),
                    largeoffsets.to(nplike),
                    slicestarts.to(nplike),
                    slicestops.to(nplike),
                    len(slicestarts),
                    missing.to(nplike),
                )
            )

            if isinstance(
                slicecontent._content,
                ak._v2.contents.listoffsetarray.ListOffsetArray,
            ):
                nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
                next = ak._v2.contents.listoffsetarray.ListOffsetArray(
                    smalloffsets, nextcontent, None, self._parameters
                )
                out = next._getitem_next_jagged(
                    smalloffsets[:-1], smalloffsets[1:], slicecontent._content, tail
                )

            else:
                out = self._getitem_next_jagged(
                    smalloffsets[:-1], smalloffsets[1:], slicecontent._content, tail
                )

            if isinstance(out, ak._v2.contents.listoffsetarray.ListOffsetArray):
                content = out._content
                missing_trim = ak._v2.index.Index64(missing[0 : largeoffsets[-1]])
                indexedoptionarray = (
                    ak._v2.contents.indexedoptionarray.IndexedOptionArray(
                        missing_trim, content, None, self._parameters
                    )
                )
                return ak._v2.contents.listoffsetarray.ListOffsetArray(
                    largeoffsets,
                    indexedoptionarray._simplify_optiontype(),
                    None,
                    self._parameters,
                )
            else:
                raise AssertionError(
                    "expected ListOffsetArray from ListArray._getitem_next_jagged, got {0}".format(
                        type(out).__name__
                    )
                )

        elif isinstance(slicecontent, ak._v2.contents.emptyarray.EmptyArray):
            return self

        else:
            raise AssertionError(
                "expected Index/IndexedOptionArray/ListOffsetArray in ListArray._getitem_next_jagged, got {0}".format(
                    type(slicecontent).__name__
                )
            )

    def _getitem_next(self, head, tail, advanced):
        nplike = self.nplike

        if head == ():
            return self

        elif isinstance(head, int):
            assert advanced is None
            nexthead, nexttail = ak._v2._slicing.headtail(tail)
            lenstarts = len(self._starts)
            nextcarry = ak._v2.index.Index64.empty(lenstarts, nplike)
            self._handle_error(
                nplike[
                    "awkward_ListArray_getitem_next_at",
                    nextcarry.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    nextcarry.to(nplike),
                    self._starts.to(nplike),
                    self._stops.to(nplike),
                    lenstarts,
                    head,
                )
            )
            nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
            return nextcontent._getitem_next(nexthead, nexttail, advanced)

        elif isinstance(head, slice):
            lenstarts = len(self._starts)

            nexthead, nexttail = ak._v2._slicing.headtail(tail)

            start, stop, step = head.start, head.stop, head.step

            step = 1 if step is None else step
            start = ak._util.kSliceNone if start is None else start
            stop = ak._util.kSliceNone if stop is None else stop

            carrylength = ak._v2.index.Index64.empty(1, nplike)
            self._handle_error(
                nplike[
                    "awkward_ListArray_getitem_next_range_carrylength",
                    carrylength.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    carrylength.to(nplike),
                    self._starts.to(nplike),
                    self._stops.to(nplike),
                    lenstarts,
                    start,
                    stop,
                    step,
                )
            )
            if self._starts.dtype == "int64":
                nextoffsets = ak._v2.index.Index64.empty(lenstarts + 1, nplike)
            elif self._starts.dtype == "int32":
                nextoffsets = ak._v2.index.Index32.empty(lenstarts + 1, nplike)
            elif self._starts.dtype == "uint32":
                nextoffsets = ak._v2.index.IndexU32.empty(lenstarts + 1, nplike)
            nextcarry = ak._v2.index.Index64.empty(carrylength[0], nplike)

            self._handle_error(
                nplike[
                    "awkward_ListArray_getitem_next_range",
                    nextoffsets.dtype.type,
                    nextcarry.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    nextoffsets.to(nplike),
                    nextcarry.to(nplike),
                    self._starts.to(nplike),
                    self._stops.to(nplike),
                    lenstarts,
                    start,
                    stop,
                    step,
                )
            )

            nextcontent = self._content._carry(nextcarry, True, NestedIndexError)

            if advanced is None or len(advanced) == 0:
                return ak._v2.contents.listoffsetarray.ListOffsetArray(
                    nextoffsets,
                    nextcontent._getitem_next(nexthead, nexttail, advanced),
                    self._identifier,
                    self._parameters,
                )
            else:
                total = ak._v2.index.Index64.empty(1, nplike)
                self._handle_error(
                    nplike[
                        "awkward_ListArray_getitem_next_range_counts",
                        total.dtype.type,
                        nextoffsets.dtype.type,
                    ](
                        total.to(nplike),
                        nextoffsets.to(nplike),
                        lenstarts,
                    )
                )

                nextadvanced = ak._v2.index.Index64.empty(total[0], nplike)
                self._handle_error(
                    nplike[
                        "awkward_ListArray_getitem_next_range_spreadadvanced",
                        nextadvanced.dtype.type,
                        advanced.dtype.type,
                        nextoffsets.dtype.type,
                    ](
                        nextadvanced.to(nplike),
                        advanced.to(nplike),
                        nextoffsets.to(nplike),
                        lenstarts,
                    )
                )
                return ak._v2.contents.listoffsetarray.ListOffsetArray(
                    nextoffsets,
                    nextcontent._getitem_next(nexthead, nexttail, nextadvanced),
                    self._identifier,
                    self._parameters,
                )

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.index.Index64):
            lenstarts = len(self._starts)

            nexthead, nexttail = ak._v2._slicing.headtail(tail)
            flathead = nplike.asarray(head.data.reshape(-1))
            regular_flathead = ak._v2.index.Index64(flathead)
            if advanced is None or len(advanced) == 0:
                nextcarry = ak._v2.index.Index64.empty(
                    lenstarts * len(flathead), nplike
                )
                nextadvanced = ak._v2.index.Index64.empty(
                    lenstarts * len(flathead), nplike
                )
                self._handle_error(
                    nplike[
                        "awkward_ListArray_getitem_next_array",
                        nextcarry.dtype.type,
                        nextadvanced.dtype.type,
                        self._starts.dtype.type,
                        self._stops.dtype.type,
                        regular_flathead.dtype.type,
                    ](
                        nextcarry.to(nplike),
                        nextadvanced.to(nplike),
                        self._starts.to(nplike),
                        self._stops.to(nplike),
                        regular_flathead.to(nplike),
                        lenstarts,
                        len(regular_flathead),
                        len(self._content),
                    ),
                    head,
                )
                nextcontent = self._content._carry(nextcarry, True, NestedIndexError)

                out = nextcontent._getitem_next(nexthead, nexttail, nextadvanced)
                if advanced is None:
                    return ak._v2._slicing.getitem_next_array_wrap(
                        out, head.metadata.get("shape", (len(head),))
                    )
                else:
                    return out

            else:
                nextcarry = ak._v2.index.Index64.empty(lenstarts, nplike)
                nextadvanced = ak._v2.index.Index64.empty(lenstarts, nplike)
                self._handle_error(
                    nplike[
                        "awkward_ListArray_getitem_next_array_advanced",
                        nextcarry.dtype.type,
                        nextadvanced.dtype.type,
                        self._starts.dtype.type,
                        self._stops.dtype.type,
                        regular_flathead.dtype.type,
                        advanced.dtype.type,
                    ](
                        nextcarry.to(nplike),
                        nextadvanced.to(nplike),
                        self._starts.to(nplike),
                        self._stops.to(nplike),
                        regular_flathead.to(nplike),
                        advanced.to(nplike),
                        lenstarts,
                        len(regular_flathead),
                        len(self._content),
                    )
                )
                nextcontent = self._content._carry(nextcarry, True, NestedIndexError)

                return nextcontent._getitem_next(nexthead, nexttail, nextadvanced)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            if advanced is not None:
                raise NestedIndexError(
                    self,
                    head,
                    "cannot mix jagged slice with NumPy-style advanced indexing",
                )
            length = len(self)
            singleoffsets = ak._v2.index.Index64(head.offsets.data)
            multistarts = ak._v2.index.Index64.empty(len(head) * length, nplike)
            multistops = ak._v2.index.Index64.empty(len(head) * length, nplike)
            nextcarry = ak._v2.index.Index64.empty(len(head) * length, nplike)

            self._handle_error(
                nplike[
                    "awkward_ListArray_getitem_jagged_expand",
                    multistarts.dtype.type,
                    multistops.dtype.type,
                    singleoffsets.dtype.type,
                    nextcarry.dtype.type,
                    self._starts.dtype.type,
                    self._stops.dtype.type,
                ](
                    multistarts.to(nplike),
                    multistops.to(nplike),
                    singleoffsets.to(nplike),
                    nextcarry.to(nplike),
                    self._starts.to(nplike),
                    self._stops.to(nplike),
                    len(head),
                    length,
                ),
            )
            carried = self._content._carry(nextcarry, True, NestedIndexError)
            down = carried._getitem_next_jagged(
                multistarts, multistops, head._content, tail
            )

            return ak._v2.contents.regulararray.RegularArray(
                down, len(head), 1, None, self._parameters
            )

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def _localindex(self, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        elif posaxis == depth + 1:
            offsets = self._compact_offsets64(True)
            innerlength = offsets[len(offsets) - 1]
            localindex = ak._v2.index.Index64.empty(innerlength, self.nplike)
            self._handle_error(
                self.nplike[
                    "awkward_ListArray_localindex",
                    localindex.dtype.type,
                    offsets.dtype.type,
                ](localindex.to(self.nplike), offsets.to(self.nplike), len(offsets) - 1)
            )
            return ak._v2.contents.listoffsetarray.ListOffsetArray(
                offsets,
                ak._v2.contents.NumpyArray(localindex),
                self._identifier,
                self._parameters,
            )
        else:
            return ak._v2.contents.listarray.ListArray(
                self._starts,
                self._stops,
                self._content._localindex(posaxis, depth + 1),
                self._identifier,
                self._parameters,
            )

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if len(self._starts) == 0:
            return ak._v2.contents.NumpyArray(self.nplike.empty(0, np.int64))

        next = self.toListOffsetArray64(True)
        out = next._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )
        return out

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        return self.toListOffsetArray64(True)._sort_next(
            negaxis,
            starts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        return ListOffsetArray._combinations(
            self, n, replacement, recordlookup, parameters, axis, depth
        )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        return self.toListOffsetArray64(True)._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

    def _validityerror(self, path):
        if len(self.stops) < len(self.starts):
            return 'at {0} ("{1}"): len(stops) < len(starts)'.format(path, type(self))
        error = self.nplike[
            "awkward_ListArray_validity", self.starts.dtype.type, self.stops.dtype.type
        ](
            self.starts.to(self.nplike),
            self.stops.to(self.nplike),
            len(self.starts),
            len(self.content),
        )
        if error.str is not None:
            if error.filename is None:
                filename = ""
            else:
                filename = " (in compiled code: " + error.filename.decode(
                    errors="surrogateescape"
                ).lstrip("\n").lstrip("(")
            message = error.str.decode(errors="surrogateescape")
            return 'at {0} ("{1}"): {2} at i={3}{4}'.format(
                path, type(self), message, error.id, filename
            )
        else:
            if (
                self.parameter("__array__") == "string"
                or self.parameter("__array__") == "bytestring"
            ):
                return ""
            else:
                return self.content.validityerror(path + ".content")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def func():
    return ak.Array([1, 2, 3, 4, 5])


generator = ak.layout.ArrayGenerator(func, (), length=5)


def test():
    hold_cache = ak._util.MappingProxy({})
    cache = ak.layout.ArrayCache(hold_cache)
    layout = ak.layout.VirtualArray(generator, cache=cache)
    array = ak.Array(layout)

    cache_2 = {}
    array_2 = ak.with_cache(array, cache_2)
    ak.materialized(array_2)

    assert len(cache_2) > 0
    assert len(cache) == 0



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
from awkward._v2.types.type import Type
from awkward._v2.forms.form import _parameters_equal


class ListType(Type):
    def __init__(self, content, parameters=None, typestr=None):
        if not isinstance(content, Type):
            raise TypeError(
                "{0} 'content' must be a Type subtype, not {1}".format(
                    type(self).__name__, repr(content)
                )
            )
        if parameters is not None and not isinstance(parameters, dict):
            raise TypeError(
                "{0} 'parameters' must be of type dict or None, not {1}".format(
                    type(self).__name__, repr(parameters)
                )
            )
        if typestr is not None and not ak._util.isstr(typestr):
            raise TypeError(
                "{0} 'typestr' must be of type string or None, not {1}".format(
                    type(self).__name__, repr(typestr)
                )
            )
        self._content = content
        self._parameters = parameters
        self._typestr = typestr

    @property
    def content(self):
        return self._content

    def __str__(self):
        if self._typestr is not None:
            out = self._typestr

        elif self.parameter("__array__") == "string":
            return "string"

        elif self.parameter("__array__") == "bytestring":
            return "bytes"

        else:
            params = self._str_parameters()
            if params is None:
                out = "var * {0}".format(str(self._content))
            else:
                out = "[var * {0}, {1}]".format(str(self._content), params)

        return self._str_categorical_begin() + out + self._str_categorical_end()

    def __repr__(self):
        args = [repr(self._content)] + self._repr_args()
        return "{0}({1})".format(type(self).__name__, ", ".join(args))

    def __eq__(self, other):
        if isinstance(other, ListType):
            return (
                self._typestr == other._typestr
                and _parameters_equal(self._parameters, other._parameters)
                and self._content == other._content
            )
        else:
            return False



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    a = ak.partition.IrregularlyPartitionedArray(
        [
            ak.Array(
                [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]
            ).layout,
            ak.Array(
                [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}, {"x": 6, "y": 6.6}]
            ).layout,
        ],
        [3, 6],
    )
    assert [a["y", i] for i in range(6)] == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
    assert [a[i, "y"] for i in range(6)] == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.emptyform import EmptyForm

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


class EmptyArray(Content):
    def __init__(self, identifier=None, parameters=None):
        self._init(identifier, parameters)

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        extra = self._repr_extra(indent + "    ")
        if len(extra) == 0:
            return indent + pre + "<EmptyArray len='0'/>" + post
        else:
            out = [indent, pre, "<EmptyArray len='0'>"]
            out.extend(extra)
            out.append("\n" + indent + "</EmptyArray>")
            out.append(post)
            return "".join(out)

    Form = EmptyForm

    @property
    def form(self):
        return self.Form(
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def typetracer(self):
        return EmptyArray(self._typetracer_identifier(), self._parameters)

    @property
    def nplike(self):
        return ak.nplike.Numpy.instance()

    @property
    def nonvirtual_nplike(self):
        return None

    def __len__(self):
        return 0

    def toNumpyArray(self, dtype, nplike=None):
        return ak._v2.contents.numpyarray.NumpyArray(
            numpy.empty(0, dtype), self._identifier, self._parameters, nplike=nplike
        )

    def _getitem_nothing(self):
        return self

    def _getitem_at(self, where):
        raise NestedIndexError(self, where, "array is empty")

    def _getitem_range(self, where):
        return self

    def _getitem_field(self, where, only_fields=()):
        raise NestedIndexError(self, where, "not an array of records")

    def _getitem_fields(self, where, only_fields=()):
        if len(where) == 0:
            return self._getitem_range(slice(0, 0))
        raise NestedIndexError(self, where, "not an array of records")

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        if len(carry) == 0:
            return self
        else:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, "array is empty")
            else:
                raise exception("array is empty")

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(head, int):
            raise NestedIndexError(self, head, "array is empty")

        elif isinstance(head, slice):
            raise NestedIndexError(self, head, "array is empty")

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.index.Index64):
            raise NestedIndexError(self, head, "array is empty")

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            raise NestedIndexError(self, head, "array is empty")

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            raise NestedIndexError(self, head, "array is empty")

        else:
            raise AssertionError(repr(head))

    def _localindex(self, axis, depth):
        return ak._v2.contents.numpyarray.NumpyArray(np.empty(0, np.int64))

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        as_numpy = ak._v2.contents.NumpyArray(self)
        return as_numpy._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        return self

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        return ak._v2.contents.emptyarray.EmptyArray(self._identifier, self._parameters)

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        as_numpy = self.toNumpyArray(reducer.preferred_dtype, nplike=parents.nplike)
        return as_numpy._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

    def _validityerror(self, path):
        return ""



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import json

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def from_iter(input):
    if input is None:
        return None

    if ak._util.isstr(input):
        return ak._v2.forms.numpyform.NumpyForm(primitive=input)

    assert isinstance(input, dict)
    has_identifier = input.get("has_identifier", input.get("has_identities", False))
    parameters = input.get("parameters", None)
    form_key = input.get("form_key", None)

    if input["class"] == "NumpyArray":
        primitive = input["primitive"]
        inner_shape = input.get("inner_shape", [])
        return ak._v2.forms.numpyform.NumpyForm(
            primitive, inner_shape, has_identifier, parameters, form_key
        )

    elif input["class"] == "EmptyArray":
        return ak._v2.forms.emptyform.EmptyForm(has_identifier, parameters, form_key)

    elif input["class"] == "RegularArray":
        return ak._v2.forms.regularform.RegularForm(
            content=from_iter(input["content"]),
            size=input["size"],
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in ("ListArray", "ListArray32", "ListArrayU32", "ListArray64"):
        return ak._v2.forms.listform.ListForm(
            starts=input["starts"],
            stops=input["stops"],
            content=from_iter(input["content"]),
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in (
        "ListOffsetArray",
        "ListOffsetArray32",
        "ListOffsetArrayU32",
        "ListOffsetArray64",
    ):
        return ak._v2.forms.listoffsetform.ListOffsetForm(
            offsets=input["offsets"],
            content=from_iter(input["content"]),
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "RecordArray":
        if isinstance(input["contents"], dict):
            contents = []
            keys = []
            for key, content in input["contents"].items():
                contents.append(from_iter(content))
                keys.append(key)
        else:
            contents = [from_iter(content) for content in input["contents"]]
            keys = None
        return ak._v2.forms.recordform.RecordForm(
            contents=contents,
            keys=keys,
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in (
        "IndexedArray",
        "IndexedArray32",
        "IndexedArrayU32",
        "IndexedArray64",
    ):
        return ak._v2.forms.indexedform.IndexedForm(
            index=input["index"],
            content=from_iter(input["content"]),
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in (
        "IndexedOptionArray",
        "IndexedOptionArray32",
        "IndexedOptionArray64",
    ):
        return ak._v2.forms.indexedoptionform.IndexedOptionForm(
            index=input["index"],
            content=from_iter(input["content"]),
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "ByteMaskedArray":
        return ak._v2.forms.bytemaskedform.ByteMaskedForm(
            mask=input["mask"],
            content=from_iter(input["content"]),
            valid_when=input["valid_when"],
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "BitMaskedArray":
        return ak._v2.forms.bitmaskedform.BitMaskedForm(
            mask=input["mask"],
            content=from_iter(input["content"]),
            valid_when=input["valid_when"],
            lsb_order=input["lsb_order"],
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "UnmaskedArray":
        return ak._v2.forms.unmaskedform.UnmaskedForm(
            content=from_iter(input["content"]),
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] in (
        "UnionArray",
        "UnionArray8_32",
        "UnionArray8_U32",
        "UnionArray8_64",
    ):
        return ak._v2.forms.unionform.UnionForm(
            tags=input["tags"],
            index=input["index"],
            contents=[from_iter(content) for content in input["contents"]],
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    elif input["class"] == "VirtualArray":
        return ak._v2.forms.virtualform.VirtualForm(
            form=from_iter(input["form"]),
            has_length=input["has_length"],
            has_identifier=has_identifier,
            parameters=parameters,
            form_key=form_key,
        )

    else:
        raise ValueError(
            "Input class: {0} was not recognised".format(repr(input["class"]))
        )


def from_json(input):
    return from_iter(json.loads(input))


def _parameters_equal(one, two):
    if one is None and two is None:
        return True
    elif one is None:
        for value in two.values():
            if value is not None:
                return False
        else:
            return True
    elif two is None:
        for value in one.values():
            if value is not None:
                return False
        else:
            return True
    else:
        keys = set(one.keys()).union(two.keys())
        for key in keys:
            if one.get(key) != two.get(key):
                return False
        else:
            return True


def _parameters_update(one, two):
    for k, v in two.items():
        if v is not None:
            one[k] = v


def nonvirtual(form):
    if isinstance(form, ak._v2.forms.virtualform.VirtualForm):
        return form.form

    else:
        return form


class Form(object):
    def _init(self, has_identifier, parameters, form_key):
        if not isinstance(has_identifier, bool):
            raise TypeError(
                "{0} 'has_identifier' must be of type bool, not {1}".format(
                    type(self).__name__, repr(has_identifier)
                )
            )
        if parameters is not None and not isinstance(parameters, dict):
            raise TypeError(
                "{0} 'parameters' must be of type dict or None, not {1}".format(
                    type(self).__name__, repr(parameters)
                )
            )
        if form_key is not None and not ak._util.isstr(form_key):
            raise TypeError(
                "{0} 'form_key' must be of type string or None, not {1}".format(
                    type(self).__name__, repr(form_key)
                )
            )

        self._has_identifier = has_identifier
        self._parameters = parameters
        self._form_key = form_key

    @property
    def has_identifier(self):
        return self._has_identifier

    @property
    def parameters(self):
        if self._parameters is None:
            self._parameters = {}
        return self._parameters

    def parameter(self, key):
        if self._parameters is None:
            return None
        else:
            return self._parameters.get(key)

    @property
    def form_key(self):
        return self._form_key

    def __str__(self):
        return json.dumps(self.tolist(verbose=False), indent=4)

    def tolist(self, verbose=True):
        return self._tolist_part(verbose, toplevel=True)

    def _tolist_extra(self, out, verbose):
        if verbose or self._has_identifier:
            out["has_identifier"] = self._has_identifier
        if verbose or (self._parameters is not None and len(self._parameters) > 0):
            out["parameters"] = self.parameters
        if verbose or self._form_key is not None:
            out["form_key"] = self._form_key
        return out

    def to_json(self):
        return json.dumps(self.tolist(verbose=True))

    def _repr_args(self):
        out = []
        if self._has_identifier is not False:
            out.append("has_identifier=" + repr(self._has_identifier))

        if self._parameters is not None and len(self._parameters) > 0:
            out.append("parameters=" + repr(self._parameters))

        if self._form_key is not None:
            out.append("form_key=" + repr(self._form_key))
        return out

    @property
    def type(self):
        return self._type({})

    def simplify_optiontype(self):
        return self

    def simplify_uniontype(self):
        return self



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import sys
import argparse
import pkg_resources

if __name__ == "__main__":
    argparser = argparse.ArgumentParser(
        description="Print out compilation arguments to use Awkward Array as a C++ dependency"
    )
    argparser.add_argument(
        "--cflags",
        action="store_true",
        help="output compiler flags and Awkward include path",
    )
    argparser.add_argument(
        "--libs", action="store_true", help="output Awkward libraries with path"
    )
    argparser.add_argument(
        "--libs-only-L", action="store_true", help="output Awkward library path"
    )
    argparser.add_argument(
        "--libs-only-l",
        action="store_true",
        help="output Awkward libraries without path",
    )
    argparser.add_argument(
        "--static-libs",
        action="store_true",
        help="output Awkward static libraries with path",
    )
    argparser.add_argument(
        "--static-libs-only-L",
        action="store_true",
        help="output Awkward static library path",
    )
    argparser.add_argument(
        "--static-libs-only-l",
        action="store_true",
        help="output Awkward static libraries without path",
    )
    argparser.add_argument(
        "--cflags-only-I", action="store_true", help="output Awkward include path"
    )
    argparser.add_argument(
        "--incdir", action="store_true", help="output Awkward include directory name"
    )
    argparser.add_argument(
        "--libdir", action="store_true", help="output Awkward library directory name"
    )

    # only used in validating the arguments
    args = argparser.parse_args()

    output = []
    incdir = pkg_resources.resource_filename("awkward", "include")
    libdir = pkg_resources.resource_filename("awkward", "")
    cpu_kernels = "awkward-cpu-kernels"
    libawkward = "awkward"

    # loop over original sys.argv to get optional arguments in order
    for arg in sys.argv:
        if arg == "--cflags":
            output.append("-std=c++11 -I{0}".format(incdir))

        if arg == "--libs":
            output.append(
                "-L{0} -l{1} -l{2} -ldl".format(libdir, libawkward, cpu_kernels)
            )

        if arg == "--libs-only-L":
            output.append("-L{0}".format(libdir))

        if arg == "--libs-only-l":
            output.append("-l{0} -l{1} -ldl".format(libawkward, cpu_kernels))

        if arg == "--static-libs":
            output.append(
                "-L{0} -l{1}-static -l{2}-static -ldl".format(
                    libdir, libawkward, cpu_kernels
                )
            )

        if arg == "--static-libs-only-L":
            output.append("-L{0}".format(libdir))

        if arg == "--static-libs-only-l":
            output.append(
                "-l{0}-static -l{1}-static -ldl".format(libawkward, cpu_kernels)
            )

        if arg == "--cflags-only-I":
            output.append("-I{0}".format(incdir))

        if arg == "--incdir":
            output.append(incdir)

        if arg == "--libdir":
            output.append(libdir)

    print(" ".join(output))  # noqa: T001



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_types_with_parameters():
    t = ak.types.UnknownType()
    assert t.parameters == {}
    t.parameters = {"__array__": ["val", "ue"]}
    assert t.parameters == {"__array__": ["val", "ue"]}
    t = ak.types.UnknownType(parameters={"__array__": ["val", "ue"]})
    assert t.parameters == {"__array__": ["val", "ue"]}

    t = ak.types.PrimitiveType("int32", parameters={"__array__": ["val", "ue"]})
    assert t.parameters == {"__array__": ["val", "ue"]}
    t = ak.types.PrimitiveType("float64", parameters={"__array__": ["val", "ue"]})
    assert t.parameters == {"__array__": ["val", "ue"]}
    t = ak.types.ArrayType(
        ak.types.PrimitiveType("int32"), 100, parameters={"__array__": ["val", "ue"]}
    )
    assert t.parameters == {"__array__": ["val", "ue"]}
    t = ak.types.ListType(
        ak.types.PrimitiveType("int32"), parameters={"__array__": ["val", "ue"]}
    )
    assert t.parameters == {"__array__": ["val", "ue"]}
    t = ak.types.RegularType(
        ak.types.PrimitiveType("int32"), 5, parameters={"__array__": ["val", "ue"]}
    )
    assert t.parameters == {"__array__": ["val", "ue"]}
    t = ak.types.OptionType(
        ak.types.PrimitiveType("int32"), parameters={"__array__": ["val", "ue"]}
    )
    assert t.parameters == {"__array__": ["val", "ue"]}
    t = ak.types.UnionType(
        (ak.types.PrimitiveType("int32"), ak.types.PrimitiveType("float64")),
        parameters={"__array__": ["val", "ue"]},
    )
    assert t.parameters == {"__array__": ["val", "ue"]}
    t = ak.types.RecordType(
        {
            "one": ak.types.PrimitiveType("int32"),
            "two": ak.types.PrimitiveType("float64"),
        },
        parameters={"__array__": ["val", "ue"]},
    )
    assert t.parameters == {"__array__": ["val", "ue"]}

    t = ak.types.UnknownType(
        parameters={"key1": ["val", "ue"], "__record__": u"one \u2192 two"}
    )
    assert t.parameters == {"__record__": u"one \u2192 two", "key1": ["val", "ue"]}

    assert t == ak.types.UnknownType(
        parameters={"__record__": u"one \u2192 two", "key1": ["val", "ue"]}
    )
    assert t != ak.types.UnknownType(parameters={"__array__": ["val", "ue"]})


def test_dress():
    class Dummy(ak.highlevel.Array):
        def __repr__(self):
            return "<Dummy {0}>".format(str(self))

    ns = {"Dummy": Dummy}

    x = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    x.setparameter("__array__", "Dummy")
    a = ak.Array(x, behavior=ns, check_valid=True)
    assert repr(a) == "<Dummy [1.1, 2.2, 3.3, 4.4, 5.5]>"

    x2 = ak.layout.ListOffsetArray64(
        ak.layout.Index64(np.array([0, 3, 3, 5], dtype=np.int64)),
        ak.layout.NumpyArray(
            np.array([1.1, 2.2, 3.3, 4.4, 5.5]), parameters={"__array__": "Dummy"}
        ),
    )
    a2 = ak.Array(x2, behavior=ns, check_valid=True)
    assert (
        repr(a2)
        == "<Array [<Dummy [1.1, 2.2, 3.3]>, ... ] type='3 * var * float64[parameters={\"__ar...'>"
    )
    assert repr(a2[0]) == "<Dummy [1.1, 2.2, 3.3]>"
    assert repr(a2[1]) == "<Dummy []>"
    assert repr(a2[2]) == "<Dummy [4.4, 5.5]>"


def test_record_name():
    typestrs = {}
    builder = ak.layout.ArrayBuilder()

    builder.beginrecord("Dummy")
    builder.field("one")
    builder.integer(1)
    builder.field("two")
    builder.real(1.1)
    builder.endrecord()

    builder.beginrecord("Dummy")
    builder.field("two")
    builder.real(2.2)
    builder.field("one")
    builder.integer(2)
    builder.endrecord()

    a = builder.snapshot()
    assert repr(a.type(typestrs)) == 'Dummy["one": int64, "two": float64]'
    assert a.type(typestrs).parameters == {"__record__": "Dummy"}


def test_builder_string():
    builder = ak.ArrayBuilder()

    builder.bytestring(b"one")
    builder.bytestring(b"two")
    builder.bytestring(b"three")

    a = builder.snapshot()
    if ak._util.py27:
        assert str(a) == "['one', 'two', 'three']"
    else:
        assert str(a) == "[b'one', b'two', b'three']"
    assert ak.to_list(a) == [b"one", b"two", b"three"]
    assert ak.to_json(a) == '["one","two","three"]'
    if ak._util.py27:
        assert repr(a) == "<Array ['one', 'two', 'three'] type='3 * bytes'>"
    else:
        assert repr(a) == "<Array [b'one', b'two', b'three'] type='3 * bytes'>"
    assert repr(ak.type(a)) == "3 * bytes"

    builder = ak.ArrayBuilder()

    builder.string("one")
    builder.string("two")
    builder.string("three")

    a = builder.snapshot()
    if ak._util.py27:
        assert str(a) == "[u'one', u'two', u'three']"
    else:
        assert str(a) == "['one', 'two', 'three']"
    assert ak.to_list(a) == ["one", "two", "three"]
    assert ak.to_json(a) == '["one","two","three"]'
    if ak._util.py27:
        assert repr(a) == "<Array [u'one', u'two', u'three'] type='3 * string'>"
    else:
        assert repr(a) == "<Array ['one', 'two', 'three'] type='3 * string'>"
    assert repr(ak.type(a)) == "3 * string"

    builder = ak.ArrayBuilder()

    builder.begin_list()
    builder.string("one")
    builder.string("two")
    builder.string("three")
    builder.end_list()

    builder.begin_list()
    builder.end_list()

    builder.begin_list()
    builder.string("four")
    builder.string("five")
    builder.end_list()

    a = builder.snapshot()
    if ak._util.py27:
        assert str(a) == "[[u'one', u'two', u'three'], [], [u'four', u'five']]"
    else:
        assert str(a) == "[['one', 'two', 'three'], [], ['four', 'five']]"
    assert ak.to_list(a) == [["one", "two", "three"], [], ["four", "five"]]
    assert ak.to_json(a) == '[["one","two","three"],[],["four","five"]]'
    assert repr(ak.type(a)) == "3 * var * string"


def test_fromiter_fromjson():
    assert ak.to_list(ak.from_iter(["one", "two", "three"])) == ["one", "two", "three"]
    assert ak.to_list(
        ak.from_iter([["one", "two", "three"], [], ["four", "five"]])
    ) == [["one", "two", "three"], [], ["four", "five"]]

    assert ak.to_list(ak.from_json('["one", "two", "three"]')) == [
        "one",
        "two",
        "three",
    ]
    assert ak.to_list(
        ak.from_json('[["one", "two", "three"], [], ["four", "five"]]')
    ) == [["one", "two", "three"], [], ["four", "five"]]
